# -*- coding: utf-8 -*-

"""Interface to openbabel."""

import logging
from pathlib import Path

try:
    import openbabel  # noqa: F401
    import openbabel.openbabel as ob
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise

logger = logging.getLogger(__name__)


class OpenBabelMixin:
    """A mixin for handling OpenBabel via its Python interface."""

    def to_OBMol(self, properties=None):
        """Return an OBMol object for the configuration, template, or subset."""
        if self.__class__.__name__ == "_Configuration":
            charge = self.charge
            spin = self.spin_multiplicity
        else:
            charge = None
            spin = None

        ob_mol = ob.OBMol()
        for atno, xyz in zip(self.atoms.atomic_numbers, self.atoms.coordinates):
            ob_atom = ob_mol.NewAtom()
            ob_atom.SetAtomicNum(atno)
            ob_atom.SetVector(*xyz)
            if charge is not None:
                ob_atom.SetFormalCharge(charge)
                charge = None
            if spin is not None:
                ob_atom.SetSpinMultiplicity(spin)
                spin = None

        # 1-based indices in ob.
        index = {j: i for i, j in enumerate(self.atoms.ids, start=1)}
        for row in self.bonds.bonds():
            ob_mol.AddBond(index[row["i"]], index[row["j"]], row["bondorder"])

        if self.__class__.__name__ == "_Configuration":
            ob_mol.SetTotalCharge(self.charge)
            if self.spin_multiplicity is None:
                n_electrons = sum(self.atoms.atomic_numbers) - self.charge
                if n_electrons % 2 == 0:
                    multiplicity = 1
                else:
                    multiplicity = 2
                self.spin_multiplicity = multiplicity
            ob_mol.SetTotalSpinMultiplicity(self.spin_multiplicity)

        # Set local radical character
        ob_mol.AssignSpinMultiplicity(True)

        # Add the net charge and spin multiplicity as properties for configurations
        pair = ob.OBPairData()

        if self.__class__.__name__ == "_Configuration":
            pair.SetAttribute("net charge")
            pair.SetValue(str(self.charge))
            ob_mol.CloneData(pair)

            pair.SetAttribute("spin multiplicity")
            pair.SetValue(str(self.spin_multiplicity))
            ob_mol.CloneData(pair)

        if properties == "all":
            data = self.properties.get("all", include_system_properties=True)
            for key, value in data.items():
                pair.SetAttribute(key)
                pair.SetValue(str(value))
                ob_mol.CloneData(pair)

                # Units, if any
                units = self.properties.units(key)
                if units is not None and units != "":
                    tmp = key.split("#", maxsplit=1)
                    if len(tmp) > 1:
                        pair.SetAttribute(tmp[0] + ",units" + "#" + tmp[1])
                    else:
                        pair.SetAttribute(key + ",units")
                    pair.SetValue(units)
                    ob_mol.CloneData(pair)
        return ob_mol

    def from_OBMol(
        self, ob_mol, properties="all", atoms=True, coordinates=True, bonds=True
    ):
        """Transform an Open Babel molecule into the current object.

        Parameters
        ----------
        rdk_mol : rdkit.chem.molecule
            The RDKit molecule object

        properties : str = "all"
            Whether to include all properties or none

        atoms : bool = True
            Recreate the atoms

        coordinates : bool = True
            Update the coordinates

        bonds : bool = True
            Recreate the bonds from the RDKit molecule

        Returns
        -------
        molsystem._Configuration
        """
        atnos = []
        Xs = []
        Ys = []
        Zs = []
        qs = []
        for ob_atom in ob.OBMolAtomIter(ob_mol):
            atno = ob_atom.GetAtomicNum()
            atnos.append(atno)
            Xs.append(ob_atom.x())
            Ys.append(ob_atom.y())
            Zs.append(ob_atom.z())
            logger.debug(f"atom {atno} {ob_atom.x()} {ob_atom.z()} {ob_atom.z()}")
            qs.append(ob_atom.GetFormalCharge())

        Is = []
        Js = []
        BondOrders = []
        for ob_bond in ob.OBMolBondIter(ob_mol):
            ob_i = ob_bond.GetBeginAtom()
            ob_j = ob_bond.GetEndAtom()
            i = ob_i.GetIdx()
            j = ob_j.GetIdx()
            bondorder = ob_bond.GetBondOrder()
            Is.append(i)
            Js.append(j)
            BondOrders.append(bondorder)
            logger.debug(f"bond {i} - {j} {bondorder}")

        if atoms:
            self.clear()

        # Get the property data, cast to correct type
        data = {}
        for item in ob_mol.GetData():
            value = item.GetValue()
            try:
                value = int(value)
            except Exception:
                try:
                    value = float(value)
                except Exception:
                    pass
            data[item.GetAttribute()] = value

        # Check for property items for charge and multiplicity
        if self.__class__.__name__ == "_Configuration":
            self.charge = ob_mol.GetTotalCharge()
            self.spin_multiplicity = ob_mol.GetTotalSpinMultiplicity()
            if "net charge" in data:
                self.charge = int(data["net charge"])
            if "spin multiplicity" in data:
                self.spin_multiplicity = int(data["spin multiplicity"])

        if atoms:
            if any([i != 0.0 for i in qs]):
                if "formal_charge" not in self.atoms:
                    self.atoms.add_attribute("formal_charge", coltype="int", default=0)
                ids = self.atoms.append(x=Xs, y=Ys, z=Zs, atno=atnos, formal_charge=qs)
            else:
                ids = self.atoms.append(x=Xs, y=Ys, z=Zs, atno=atnos)
        else:
            ids = self.atoms.ids

            if coordinates:
                xyz = [[x, y, z] for x, y, z in zip(Xs, Ys, Zs)]
                self.atoms.coordinates = xyz

        if atoms or bonds:
            i = [ids[x - 1] for x in Is]
            j = [ids[x - 1] for x in Js]
            self.bonds.append(i=i, j=j, bondorder=BondOrders)

        # Record any properties in the database if desired
        if properties == "all":
            for key, value in data.items():
                if ",units" not in key and key not in [
                    "net charge",
                    "spin multiplicity",
                ]:
                    if not self.properties.exists(key):
                        tmp = key.split("#", maxsplit=1)
                        if len(tmp) > 1:
                            units_key = tmp[0] + ",units" + "#" + tmp[1]
                        else:
                            units_key = key + ",units"
                        _type = value.__class__.__name__
                        if units_key in data:
                            units = data[units_key]
                            self.properties.add(key, _type, units=units)
                        else:
                            self.properties.add(key, _type)
                    self.properties.put(key, value)
        return self

    def coordinates_from_OBMol(self, ob_mol):
        """Update the coordinates from an Open Babel molecule."""
        XYZs = []
        for ob_atom in ob.OBMolAtomIter(ob_mol):
            XYZs.append([ob_atom.x(), ob_atom.y(), ob_atom.z()])

        self.coordinates = XYZs

    def coordinates_to_OBMol(self, ob_mol):
        """Update the coordinates of an Open Babel molecule from the configuration."""
        XYZs = self.coordinates
        for XYZ, ob_atom in zip(XYZs, ob.OBMolAtomIter(ob_mol)):
            ob_atom.SetVector(*XYZ)

    def find_substructures(self, template):
        """Find the substructures matching the template.

        Parameters
        ----------
        template : str, _Configuration, _Template, or _Subset
            The template, which may be a SMARTS string, or a molecular object.

        Returns
        -------
        [[int]]
            Lists of atom ids for matches.
        """
        if isinstance(template, str):
            smarts = template
        else:
            smarts = self.smiles

        ob_mol = self.to_OBMol()

        pattern = ob.OBSmartsPattern()
        pattern.Init(smarts)
        pattern.Match(ob_mol)
        maplist = pattern.GetUMapList()

        return [x for x in maplist]

    def from_sdf_text(self, text):
        """Get the text of an SDF file for the configuration.

        Parameters
        ----------
        text : str
            The text of an SDF file

        Returns
        -------
        (str, str)
            system name, configuration name
        """
        obConversion = ob.OBConversion()
        obConversion.SetInFormat("sdf")
        obMol = ob.OBMol()
        obConversion.ReadString(obMol, text)

        self.from_OBMol(obMol)

        # See if the system and configuration names are encoded in the tit
        result = (None, None)
        title = obMol.GetTitle()
        if "SEAMM=" in title:
            for tmp in title.split("|"):
                if "SEAMM=" in tmp and "/" in tmp:
                    sysname, confname = tmp.split("=", 1)[1].split("/")
                    sysname = sysname.strip()
                    confname = confname.strip()
                    result = (sysname, confname)
        return result

    def from_sdf(self, path):
        """Directly read an SDF file for the configuration.

        Parameters
        ----------
        path : pathlib.Path or str
            The path or name of the file to write.

        Returns
        -------
        (str, str)
            system name, configuration name
        """
        path = Path(path)
        text = path.read_text()

        return self.from_sdf_text(text)

    def to_sdf_text(self):
        """Get the text of an SDF file for the configuration.

        Returns
        -------
        str
            The text of the SDF file
        """
        obConversion = ob.OBConversion()
        obConversion.SetOutFormat("sdf")
        obMol = self.to_OBMol(properties="all")
        title = f"SEAMM={self.system.name}/{self.name}"
        obMol.SetTitle(title)
        text = obConversion.WriteString(obMol)

        return text

    def to_sdf(self, path):
        """Directly write an SDF file for the configuration.

        Parameters
        ----------
        path : pathlib.Path or str
            The path or name of the file to write.

        Returns
        -------
        str
            The text of the SDF file
        """
        text = self.to_sdf_text()

        path = Path(path)
        path.write_text(text)

        return text



# -*- coding: utf-8 -*-

"""Interface to RDKit."""

import logging
import pprint


try:
    from rdkit import Chem
except ModuleNotFoundError:
    print(
        "Please install rdkit using conda:\n" "     conda install -c conda-forge rdkit"
    )
    raise

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# Valence
valence = {
    1: 1,
    2: 0,
    3: 1,
    4: 2,
    5: 3,
    6: 4,
    7: 3,
    8: 2,
    9: 1,
    10: 0,
    11: 1,
    12: 2,
    13: 3,
    14: 4,
    15: 3,
    16: 2,
    17: 1,
    18: 0,
    92: 2,
}


class RDKitMixin:
    """A mixin for handling RDKit via its Python interface."""

    def to_RDKMol(self, properties=None):
        """Return an RDKMol object for the configuration, template, or subset."""
        index = {}
        indices = []
        rdk_mol = Chem.RWMol()
        for atno, _id in zip(self.atoms.atomic_numbers, self.atoms.ids):
            idx = rdk_mol.AddAtom(Chem.Atom(atno))
            index[_id] = idx
            indices.append(idx)

        bond_types = {
            1: Chem.BondType.SINGLE,
            2: Chem.BondType.DOUBLE,
            3: Chem.BondType.TRIPLE,
            5: Chem.BondType.AROMATIC,
        }
        for row in self.bonds.bonds():
            rdk_mol.AddBond(
                index[row["i"]], index[row["j"]], bond_types[row["bondorder"]]
            )

        # Check for NH4+ type groups and set their charge
        rdk_mol.UpdatePropertyCache(strict=False)
        for at in rdk_mol.GetAtoms():
            atno = at.GetAtomicNum()
            if atno in valence:
                charge = at.GetExplicitValence() - valence[atno]
                if charge != 0 and at.GetFormalCharge() == 0:
                    at.SetFormalCharge(charge)

        natom = self.atoms.n_atoms
        conf = Chem.Conformer(natom)
        for idx, xyz in zip(indices, self.atoms.coordinates):
            conf.SetAtomPosition(idx, xyz)

        rdk_mol.AddConformer(conf)
        try:
            Chem.SanitizeMol(rdk_mol)
        except Chem.KekulizeException as e:
            logger.warning(f"Kekulization failed: {e}")
            flags = (
                Chem.SanitizeFlags.SANITIZE_ALL ^ Chem.SanitizeFlags.SANITIZE_KEKULIZE
            )
            Chem.rdmolops.SanitizeMol(rdk_mol, sanitizeOps=flags)

        # Add the net charge and spin multiplicity as properties for configurations
        if self.__class__.__name__ == "_Configuration":
            rdk_mol.SetIntProp("net charge", self.charge)
            rdk_mol.SetIntProp("spin multiplicity", self.spin_multiplicity)

        if properties == "all":
            data = self.properties.get("all", include_system_properties=True)
            for key, value in data.items():
                if isinstance(value, int):
                    rdk_mol.SetIntProp(key, value)
                elif isinstance(value, float):
                    rdk_mol.SetDoubleProp(key, value)
                else:
                    rdk_mol.SetProp(key, str(value))

                # Units, if any
                units = self.properties.units(key)
                if units is not None and units != "":
                    tmp = key.split("#", maxsplit=1)
                    if len(tmp) > 1:
                        rdk_mol.SetProp(tmp[0] + ",units" + "#" + tmp[1], units)
                    else:
                        rdk_mol.SetProp(key + ",units", units)

        return rdk_mol

    def from_RDKMol(
        self, rdk_mol, properties="all", atoms=True, coordinates=True, bonds=True
    ):
        """Transform an RDKit molecule into the current object.

        Parameters
        ----------
        rdk_mol : rdkit.chem.molecule
            The RDKit molecule object

        properties : str = "all"
            Whether to include all properties or none

        atoms : bool = True
            Recreate the atoms

        coordinates : bool = True
            Update the coordinates

        bonds : bool = True
            Recreate the bonds from the RDKit molecule

        Returns
        -------
        molsystem._Configuration
        """
        atnos = []
        for rdk_atom in rdk_mol.GetAtoms():
            atnos.append(rdk_atom.GetAtomicNum())
            logger.debug(f"atom {atnos}")

        # TODO: Generalize to handling multiple conformers
        # in a rdk_mol object, if necessary
        Xs = []
        Ys = []
        Zs = []
        for rdk_conf in rdk_mol.GetConformers():
            for atom_coords in rdk_conf.GetPositions():
                Xs.append(atom_coords[0])
                Ys.append(atom_coords[1])
                Zs.append(atom_coords[2])
                logger.debug(f"{atom_coords[0]} {atom_coords[1]} {atom_coords[2]}")

        Is = []
        Js = []
        BondOrders = []
        bond_types = {
            Chem.BondType.SINGLE: 1,
            Chem.BondType.DOUBLE: 2,
            Chem.BondType.TRIPLE: 3,
            Chem.BondType.AROMATIC: 5,
        }
        for rdk_bond in rdk_mol.GetBonds():
            i = rdk_bond.GetBeginAtom().GetIdx()
            j = rdk_bond.GetEndAtom().GetIdx()
            bondorder = bond_types[rdk_bond.GetBondType()]
            Is.append(i)
            Js.append(j)
            BondOrders.append(bondorder)
            logger.debug(f"bond {i} - {j} {bondorder}")

        if atoms:
            self.clear()
            ids = self.atoms.append(x=Xs, y=Ys, z=Zs, atno=atnos)
        else:
            ids = self.atoms.ids

            if coordinates:
                xyz = [[x, y, z] for x, y, z in zip(Xs, Ys, Zs)]
                self.atoms.coordinates = xyz

        if atoms or bonds:
            i = [ids[x] for x in Is]
            j = [ids[x] for x in Js]
            self.bonds.append(i=i, j=j, bondorder=BondOrders)

        data = rdk_mol.GetPropsAsDict()
        if self.__class__.__name__ == "_Configuration":
            # Check for property items for charge and multiplicity
            if "net charge" in data:
                self.charge = int(data["net charge"])
            if "spin multiplicity" in data:
                self.spin_multiplicity = int(data["spin multiplicity"])

        # Record any properties in the database if desired
        if properties == "all":
            for key, value in data.items():
                if ",units" not in key and key not in [
                    "net charge",
                    "spin multiplicity",
                ]:
                    if not self.properties.exists(key):
                        tmp = key.split("#", maxsplit=1)
                        if len(tmp) > 1:
                            units_key = tmp[0] + ",units" + "#" + tmp[1]
                        else:
                            units_key = key + ",units"
                        _type = value.__class__.__name__
                        if units_key in data:
                            units = data[units_key]
                            self.properties.add(key, _type, units=units)
                        else:
                            self.properties.add(key, _type)
                    self.properties.put(key, value)
        return self

    def debug_print(self):
        pprint.pprint(str(self.atoms._atom_table))
        pprint.pprint(str(self.atoms._coordinates_table))
        pprint.pprint(str(self.bonds))



# -*- coding: utf-8 -*-

"""Functions for handling PubChem"""

import logging
from urllib.parse import quote as url_quote

import requests

logger = logging.getLogger(__name__)
pug_url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug"

try:
    import pubchempy as pcp
except ModuleNotFoundError:
    print(
        "To use PubChem, please install pubchempy using conda:\n"
        "     conda install -c conda-forge pubchempy"
    )
    logger.warning(
        "To use PubChem, please install pubchempy using conda:\n"
        "     conda install -c conda-forge pubchempy"
    )


class PubChemMixin:
    """A mixin for handling the PubChem database."""

    @property
    def PC_cid(self):
        """Return the PubChem CID for this structure, or None."""
        inchi = self.to_inchi()
        results = pcp.get_compounds(inchi, "inchi")

        if len(results) == 0:
            return None

        if len(results) >= 1:
            logger.info(f"PubChem search returned more than one hit of {inchi}")

        result = results[0].to_dict()
        if "cid" in result:
            return result["cid"]

        return None

    def PC_from_cid(self, cid, fallback=None):
        """Create the configuration from the PubChem 3-D structure, if available.

        Parameters
        ----------
        cid : int
            The PubChem CID.

        fallback : str
            A fallback SMILES, InChI, etc. to use if PubChem fails
        """
        response = requests.get(f"{pug_url}/compound/cid/{cid}/SDF")
        if response.status_code == 200:
            self.from_sdf_text(response.text)
            return

        # An error!
        if fallback is None:
            raise RuntimeError(f"No 3-D structure available for {cid}")

        # See what the fallback is
        if len(fallback) == 27:
            tmp = fallback.split("-")
            if (
                len(tmp) == 3
                and len(tmp[0]) == 14
                and len(tmp[1]) == 10
                and len(tmp[2]) == 1
            ):
                self.from_inchikey(fallback)
                return
        if fallback[0:7] == "InChI=":
            self.from_inchi(fallback)
            return
        else:
            self.from_smiles(fallback)
            return

    def PC_from_identifier(self, identifier, namespace="detect", fallback=None):
        """Create the configuration from the PubChem 3-D structure, if available.

        Parameters
        ----------
        identifier : int or str
            The PubChem identifier
        namespace : str
            The PubChem namespace: cid, name, smiles, inchi, inchikey
        fallback : str
            A fallback SMILES, InChI, etc. to use if PubChem fails
        """
        if namespace == "detect":
            # Work through the possibilities
            if len(identifier) == 27:
                tmp = identifier.split("-")
                if (
                    len(tmp) == 3
                    and len(tmp[0]) == 14
                    and len(tmp[1]) == 10
                    and len(tmp[2]) == 1
                ):
                    namespace = ["inchikey"]
            elif identifier[0:7] == "InChI=":
                self.from_inchi(identifier)
                namespaces = ["inchi"]
            else:
                namespaces = ["name", "smiles"]
        else:
            namespaces = [namespace]

        for namespace in namespaces:
            response = requests.get(
                f"{pug_url}/compound/{namespace}/{url_quote(identifier)}/"
                "SDF?record_type=3d"
            )
            if response.status_code == 200:
                self.from_sdf_text(response.text)
                return

        # An error!
        if fallback is None:
            raise RuntimeError(f"No 3-D structure available for {identifier}")

        # See what the fallback is
        if len(fallback) == 27:
            tmp = fallback.split("-")
            if (
                len(tmp) == 3
                and len(tmp[0]) == 14
                and len(tmp[1]) == 10
                and len(tmp[2]) == 1
            ):
                self.from_inchikey(fallback)
                return
        if fallback[0:7] == "InChI=":
            self.from_inchi(fallback)
            return
        else:
            self.from_smiles(fallback)
            return

    def PC_iupac_name(self, fallback=None):
        """Return the IUPAC name for this structure, or None.

        Parameters
        ----------
        fallback : str
            A name to return if PubChem doesn't have a name
        """
        inchi = self.to_inchi()
        results = pcp.get_compounds(inchi, "inchi")

        if len(results) == 0:
            return fallback

        if len(results) >= 1:
            logger.info(f"PubChem search returned more than one hit of {inchi}")

        result = results[0].to_dict()
        if "iupac_name" in result:
            return result["iupac_name"]

        return fallback



# -*- coding: utf-8 -*-

"""Property methods for systems."""

import logging
import pprint  # noqa: F401

logger = logging.getLogger(__name__)


class _SystemProperties(object):
    """A class for handling the properties of a system in the star schema."""

    def __init__(self, system):
        self._system = system
        self._sid = system.id
        self._properties = None

    @property
    def system(self):
        """The system these properties are connected with."""
        return self._system

    @property
    def properties(self):
        """The general properties object."""
        if self._properties is None:
            self._properties = self.system.system_db.properties
        return self._properties

    @property
    def standard_properties(self):
        """A thin wrapper of the _Properties method."""
        return self.properties.standard_properties()

    def add(self, name, _type="float", units=None, description="", noerror=False):
        """A thin wrapper of the _Properties method."""
        return self.properties.add(
            name, _type=_type, units=units, description=description, noerror=noerror
        )

    def description(self, _property):
        """The description of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The description of the property.
        """
        return self.properties.description(_property)

    def exists(self, name):
        """A thin wrapper of the _Properties method."""
        return self.properties.exists(name)

    def get(self, _property="all", include_configuration_properties=False):
        """Get the given property value for this system.

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        int, float, or str
            The value of the property.
        include_configuration_properties : bool=False
            Whether to include properties from any configuration in the system.
        """
        return self.properties.get_for_system(
            self._sid,
            _property,
            include_configuration_properties=include_configuration_properties,
        )

    def id(self, name):
        """The id for a property

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        int
            The database id for the property.
        """
        return self.properties.id(name)

    def known_properties(self):
        """List the known properties."""
        return self.properties.known_properties()

    def list(self):
        """Return all of the property names for the current system.
        Returns
        -------
        [str]
            The names of the properties for the system.
        """
        sql = "SELECT name FROM property WHERE id IN ("
        sql += "SELECT property FROM float_data WHERE system = ?"
        sql += "UNION SELECT property FROM int_data WHERE system = ?"
        sql += "UNION SELECT property FROM str_data WHERE system = ?"
        sql += ")"

        self.properties.cursor.execute(sql, (self._sid, self._sid, self._sid))
        return [row[0] for row in self.properties.cursor]

    def list_ids(self):
        """Return all of the property ids for the current system.
        Returns
        -------
        [int]
            The names of the properties for the system.
        """
        sql = "SELECT property FROM float_data WHERE system = ?"
        sql += "UNION SELECT property FROM int_data WHERE system = ?"
        sql += "UNION SELECT property FROM str_data WHERE system = ?"

        self.properties.cursor.execute(sql, (self._sid, self._sid, self._sid))
        return [row[0] for row in self.properties.cursor]

    def metadata(self, _property):
        """The metadata for a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str, str, str, str
            The name, type, units, and description of the property
        """
        return self.properties.metadata(_property)

    def name(self, pid):
        """The name of a property

        Parameters
        ----------
        pid : int
            The id of the property.

        Returns
        -------
        str
            The name of the property.
        """
        return self.property.name(pid)

    def put(self, _property, value):
        """Store the given property value for this system.

        Parameters
        ----------
        _property : int or str
            The id or name of the property.
        value : int, float, or str
            The value to store.
        """
        self.properties.put_for_system(self._sid, _property, value)

    def type(self, _property):
        """The type of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The type of the property.
        """
        return self.properties.type(_property)

    def units(self, _property):
        """The unit string of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The units string for the property.
        """
        return self.properties.units(_property)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401

"""Tests for handling PubChem."""


def test_to_smiles(AceticAcid):
    """Get the name from a system"""
    correct = "acetic acid"
    name = AceticAcid.PC_iupac_name()

    if name != correct:
        print(name)
    assert name == correct


def test_from_cid(configuration):
    """Get the system from a PubChem cid"""
    correct = "(Z)-but-2-ene"
    cid = 5287573
    configuration.PC_from_cid(cid)

    new_cid = configuration.PC_cid
    name = configuration.PC_iupac_name()

    if new_cid != cid:
        print(new_cid)
    assert new_cid == cid

    if name != correct:
        print(name)
    assert name == correct


def test_from_name(configuration):
    """Get the system from a name"""
    correct = "(Z)-but-2-ene"
    cid = 5287573
    configuration.PC_from_identifier("cis-2-butene")

    new_cid = configuration.PC_cid
    name = configuration.PC_iupac_name()

    if new_cid != cid:
        print(new_cid)
    assert new_cid == cid

    if name != correct:
        print(name)
    assert name == correct


def test_coordinates_from_name(configuration):
    """Get the system from a name"""
    correct = [
        [0.0, 0.0, 0.0],
        [-0.4417, 0.2906, 0.8711],
        [0.7256, 0.6896, -0.1907],
        [0.4875, -0.8701, 0.2089],
    ]

    configuration.PC_from_identifier("ammonia")

    coordinates = configuration.coordinates
    if coordinates != correct:
        print(coordinates)
    assert coordinates == correct


def test_coordinates_from_smiles(configuration):
    """Get the system from SMILES"""
    correct = [
        [0.0, 0.0, 0.0],
        [-0.4417, 0.2906, 0.8711],
        [0.7256, 0.6896, -0.1907],
        [0.4875, -0.8701, 0.2089],
    ]

    configuration.PC_from_identifier("N", namespace="smiles")

    coordinates = configuration.coordinates
    if coordinates != correct:
        print(coordinates)
    assert coordinates == correct



# -*- coding: utf-8 -*-

"""Functions for handling InChI"""

import logging
import requests

try:
    from openbabel import openbabel as OB
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise
try:
    from rdkit import Chem
    from rdkit.Chem import AllChem
except ModuleNotFoundError:
    print("Please install RDKit using conda:\n     conda install -c conda-forge rdkit")
    raise

logger = logging.getLogger(__name__)


class InChIMixin:
    """A mixin for handling InChI."""

    @property
    def inchi(self):
        """Return the InChI string for this object."""
        return self.to_inchi()

    @property
    def inchikey(self):
        """Return the InChIKey string for this object."""
        return self.to_inchi(key=True)

    def to_inchi(self, key=False, openbabel=False):
        """Create the InChI string from the system.

        Parameters
        ----------
        key : bool = False
            Whether to create the InChIKey
        openbabel : bool = False
            Whether to use OpenBabel rather than default of RDkit

        Returns
        -------
        str
            The InChI string, or (InChI, name) if the rname is requested
        """
        logger.info("to_inchi")

        if openbabel:
            obConversion = OB.OBConversion()
            if key:
                obConversion.SetOutFormat("inchikey")
            else:
                obConversion.SetOutFormat("inchi")

            mol = self.to_OBMol()

            inchi = obConversion.WriteString(mol)
        else:
            mol = self.to_RDKMol()
            if key:
                inchi = Chem.inchi.MolToInchiKey(mol)
            else:
                inchi = Chem.inchi.MolToInchi(mol)

        logger.info(f"inchi = '{inchi}'")

        return inchi.strip()

    def from_inchi(self, inchi, name=None, reorient=True, openbabel=True):
        """Create the system from a InChI string.

        Parameters
        ----------
        inchi : str
            The InChI string
        name : str = None
            The name of the molecule
        reorient : bool = True
            Whether to reorient to the standard orientation
        openbabel : bool = False
            Whether to use Openbabel rather than default of RDKit

        Returns
        -------
        None
        """

        save = self.name

        if openbabel:
            obConversion = OB.OBConversion()
            obConversion.SetInAndOutFormats("inchi", "mdl")
            mol = OB.OBMol()
            obConversion.ReadString(mol, inchi)

            # Add hydrogens
            mol.AddHydrogens()

            # Get coordinates for a 3-D structure
            builder = OB.OBBuilder()
            builder.Build(mol)

            self.from_OBMol(mol)
        else:
            raise NotImplementedError("RDKit from InChI doesn't seem to work!")
            mol = Chem.MolFromInchi(
                inchi, sanitize=False, removeHs=False, treatWarningAsError=True
            )
            mol = Chem.AddHs(mol)
            AllChem.EmbedMolecule(mol)
            self.from_RDKMol(mol)

        # Rotate to standard orientation
        rdkMol = self.to_RDKMol()
        rdkConf = rdkMol.GetConformers()[0]
        Chem.rdMolTransforms.CanonicalizeConformer(rdkConf)
        self.from_RDKMol(rdkMol)

        if name is not None:
            self.name = name
        else:
            self.name = save

    def from_inchikey(self, inchikey, name=None, reorient=True):
        """Create the system from an InChIKey string.

        Parameters
        ----------
        inchikey : str
            The InChIKey string
        name : str = None
            The name of the molecule
        reorient : bool = True
            Whether to reorient to the standard orientation

        Returns
        -------
        None
        """
        inchi = self._get_inchi(inchikey)
        self.from_inchi(inchi, reorient=reorient)

    def _get_inchi(self, inchikey):
        """Get the InChI from PubChem given the InChIKey."""
        url = (
            f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/inchikey/{inchikey}"
            "/property/InChI/JSON"
        )
        r = requests.get(url)
        result = r.json()
        if "Fault" in result:
            raise RuntimeError(f"InChIKey '{inchikey}' not found in PubChem.")

        inchi = set()
        for properties in result["PropertyTable"]["Properties"]:
            if "InChI" in properties:
                inchi.add(properties["InChI"])
        if len(inchi) == 1:
            return inchi.pop()
        else:
            return [*inchi]



# -*- coding: utf-8 -*-

"""Alignment methods for configurations"""

import logging

try:
    import openbabel  # noqa: F401
    import openbabel.openbabel as ob
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise

logger = logging.getLogger(__name__)


class AlignMixin:
    """A mixin for handling alignment of configuration."""

    def RMSD(self, other, include_h=True, symmetry=False):
        """Compute the RMSD between configurations.

        Parameters
        ----------
        other : Configuration or iterable of Configurations
            A single configuration or e.g. list of configurations to match

        include_h : bool = True
            Whether to include hydrogen atoms in the RMSD

        symmetry : bool = False
            Whether to detect symmetric flips. Note this requires a lot of memory for
            larger systems.

        Returns
        -------
        float or [float]
            The RMS between the current configuration and the target(s).
        """
        from .configuration import _Configuration

        align = ob.OBAlign(include_h, symmetry)
        align.SetRefMol(self.to_OBMol())

        if isinstance(other, _Configuration):
            align.SetTargetMol(other.to_OBMol())
            align.Align()
            return align.GetRMSD()
        elif isinstance(other, ob.OBMol):
            align.SetTargetMol(other)
            align.Align()
            return align.GetRMSD()

        result = []
        try:
            for target in other:
                if isinstance(target, _Configuration):
                    align.SetTargetMol(target.to_OBMol())
                    align.Align()
                    result.append(align.GetRMSD())
                elif isinstance(target, ob.OBMol):
                    align.SetTargetMol(target)
                    align.Align()
                    result.append(align.GetRMSD())
                else:
                    raise TypeError(f"RMSD can't handle {type(target)}.")
        except TypeError:
            raise
        except Exception as e:
            logger.error(f"Configuration.RMSD {e}")
            raise
        return result

    def align(self, other, include_h=True, symmetry=False):
        """Align configurations.

        Parameters
        ----------
        other : Configuration or iterable of Configurations
            A single configuration or e.g. list of configurations to match

        include_h : bool = True
            Whether to include hydrogen atoms in the RMSD

        symmetry : bool = False
            Whether to detect symmetric flips. Note this requires a lot of memory for
            larger systems.

        Returns
        -------
        float or [float]
            The RMS between the current configuration and the target(s).
        """
        from .configuration import _Configuration

        align = ob.OBAlign(include_h, symmetry)
        align.SetRefMol(self.to_OBMol())

        if isinstance(other, _Configuration):
            mol = other.to_OBMol()
            align.SetTargetMol(mol)
            align.Align()
            align.UpdateCoords(mol)
            other.coordinates_from_OBMol(mol)
            return align.GetRMSD()
        elif isinstance(other, ob.OBMol):
            align.SetTargetMol(other)
            align.Align()
            align.UpdateCoords(other)
            return align.GetRMSD()

        result = []
        try:
            for target in other:
                if isinstance(target, _Configuration):
                    mol = target.to_OBMol()
                    align.SetTargetMol(mol)
                    align.Align()
                    align.UpdateCoords(mol)
                    target.coordinates_from_OBMol(mol)
                    result.append(align.GetRMSD())
                elif isinstance(target, ob.OBMol):
                    align.SetTargetMol(target)
                    align.Align()
                    align.UpdateCoords(target)
                    result.append(align.GetRMSD())
                else:
                    raise TypeError(f"RMSD can't handle {type(target)}.")
        except TypeError:
            raise
        except Exception as e:
            logger.error(f"Configuration.RMSD {e}")
            raise
        return result



# -*- coding: utf-8 -*-

"""Interface to qcschema."""

import json
import logging

from seamm_util import Q_

logger = logging.getLogger(__name__)


class QCSchemaMixin:
    """A mixin for handling QCSchema."""

    def to_qcschema_dict(self, properties=None):
        """Create a dictionary compliant with QCSchema."""
        result = {
            "schema_name": "qcschema_molecule",
            "schema_version": 2,
        }

        # Symbols and coordinates (in Bohr)
        result["symbols"] = [*self.atoms.symbols]
        factor = Q_(1.0, "Å").m_as("a_0")
        xyz = []
        # round() below helps tests work across platforms. 9 digits are enough!
        for row in self.atoms.get_coordinates(fractionals=False):
            for val in row:
                xyz.append(round(val * factor, 9))
        result["geometry"] = xyz

        # Charge and multiplicity
        result["molecular_charge"] = self.charge
        result["molecular_multiplicity"] = self.spin_multiplicity

        # Bonds, if any
        bonds = []
        index = {j: i for i, j in zip(range(self.n_atoms), self.atoms.ids)}
        for row in self.bonds.bonds():
            bonds.append((index[row["i"]], index[row["j"]], row["bondorder"]))
        if len(bonds) > 0:
            result["connectivity"] = bonds

        # Molecules (fragments in QCSchema speak)
        result["fragments"] = self.find_molecules(as_indices=True)

        result["name"] = f"{self.system.name} / {self.name}"
        if "name" in self.atoms:
            result["atom_labels"] = self.atoms.get_column_data("name")

        return result

    def to_qcschema_json(self):
        """Create the QCSchema JSON for the molecule."""
        data = self.to_qcschema_dict()
        return json.dumps(data)

    def from_qcschema_dict(self, data):
        """Reset the molecule from the QCSchema data."""
        self.clear()
        self.periodicity = 0

        symbols = data["symbols"]
        factor = Q_(1.0, "a_0").m_as("Å")
        Xs = []
        Ys = []
        Zs = []
        for x, y, z in zip(*[iter(data["geometry"])] * 3):
            Xs.append(x * factor)
            Ys.append(y * factor)
            Zs.append(z * factor)

        ids = self.atoms.append(x=Xs, y=Ys, z=Zs, symbol=symbols)

        if "atom_labels" in data:
            if "name" not in self.atoms:
                self.atoms.add_attribute("name", values=data["atom_labels"])
            else:
                self.atoms["name"] = data["atom_labels"]

        if "connectivity" in data and len(data["connectivity"]) > 0:
            Is = []
            Js = []
            orders = []
            for i, j, order in data["connectivity"]:
                Is.append(i)
                Js.append(j)
                orders.append(order)

            i = [ids[x - 1] for x in Is]
            j = [ids[x - 1] for x in Js]
            self.bonds.append(i=i, j=j, bondorder=orders)

    def from_qcschema_json(self, json_data):
        """Reset the molecule from the QCSchema JSON."""
        data = json.loads(json_data)
        self.from_qcschema_dict(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest  # noqa: F401

"""Tests for handling InChI."""


def test_to_inchi(AceticAcid):
    """Create a InChI string from a system"""
    correct = "InChI=1S/C2H4O2/c1-2(3)4/h1H3,(H,3,4)"
    inchi = AceticAcid.to_inchi()

    if inchi != correct:
        print(inchi)
    assert inchi == correct


def test_to_inchikey(AceticAcid):
    """Create a InChIKey string from a system"""
    correct = "QTBSBXVTEAMEQO-UHFFFAOYSA-N"
    inchi = AceticAcid.to_inchi(key=True)

    if inchi != correct:
        print(inchi)
    assert inchi == correct


def test_to_inchi_with_name(AceticAcid):
    """Create a InChI string from a system, adding the name"""
    correct = "InChI=1S/C2H4O2/c1-2(3)4/h1H3,(H,3,4)"
    inchi = AceticAcid.inchi

    if inchi != correct:
        print(inchi)
    assert inchi == correct


def test_several_molecules(CH3COOH_3H2O):
    """System with acetic acid and 3 waters"""
    system = CH3COOH_3H2O
    correct = "InChI=1S/C2H4O2.3H2O/c1-2(3)4;;;/h1H3,(H,3,4);3*1H2"

    inchi = system.to_inchi()

    if inchi != correct:
        print(inchi)
    assert inchi == correct


def test_from_inchi(configuration):
    """Create a configuration from InChI"""
    correct = "CC(=O)O"
    configuration.from_inchi(
        "InChI=1S/C2H4O2/c1-2(3)4/h1H3,(H,3,4)", name="acetic acid"
    )
    result = configuration.to_smiles(canonical=True)

    if result != correct:
        print(result)

    assert result == correct


def test_from_inchikey(configuration):
    """Create a configuration from an InChIKey"""
    correct = "CC(=O)O"

    configuration.from_inchikey("QTBSBXVTEAMEQO-UHFFFAOYSA-N")
    result = configuration.to_smiles(canonical=True)

    if result != correct:
        print(result)

    assert result == correct



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for configurations."""

import pytest  # noqa: F401


def test_formula(AceticAcid):
    """Test the formula generation."""
    formula, empirical_formula, Z = AceticAcid.formula
    assert formula == "C2 H4 O2"
    assert empirical_formula == "C H2 O"
    assert Z == 2


def test_formula_periodic(vanadium):
    """Test the formula generation."""
    formula, empirical_formula, Z = vanadium.formula
    assert "".join(formula) == "V2"
    assert "".join(empirical_formula) == "V"
    assert Z == 2


def test_density(vanadium):
    """Test the density, and implicitly the mass and volume."""

    assert abs(vanadium.density - 6.0817308915133) < 1.0e-06


def test_copy(H2O):
    """Test copying and changing a configuration."""
    original = H2O
    coordinates = [
        [0.0, 0.0, 0.0],
        [0.0, 0.8, -0.8],
        [0.0, -0.8, -0.8],
    ]
    copy = original.system.copy_configuration(original, name="Copy")

    assert original.coordinates == copy.coordinates

    copy.coordinates = coordinates

    assert original.coordinates != copy.coordinates


def test_copy_periodic(vanadium):
    """Test copying and changing a configuration."""
    original = vanadium

    copy = original.system.copy_configuration(original, name="Copy")

    assert original.cell.parameters == copy.cell.parameters

    copy.cell.parameters = [3.03, 3.03, 3.03, 90, 90, 91]

    assert original.cell.parameters != copy.cell.parameters



# -*- coding: utf-8 -*-

"""Functions for handling SMILES"""

import logging

try:
    from openeye import oechem
except ModuleNotFoundError:
    oechem_available = False
    oechem_licensed = False
else:
    oechem_available = True
    oechem_licensed = oechem.OEChemIsLicensed()

try:
    from openeye import oeomega
except ModuleNotFoundError:
    oeomega_available = False
    oeomega_licensed = False
else:
    oeomega_available = True
    oeomega_licensed = oeomega.OEOmegaIsLicensed()

try:
    from openbabel import openbabel
except ModuleNotFoundError:
    print(
        "Please install openbabel using conda:\n"
        "     conda install -c conda-forge openbabel"
    )
    raise
try:
    from rdkit import Chem
    from rdkit.Chem import AllChem
except ModuleNotFoundError:
    print("Please install RDKit using conda:\n     conda install -c conda-forge rdkit")
    raise

logger = logging.getLogger(__name__)


def check_openeye_license():
    if not oechem_available:
        raise RuntimeError(
            "OpenEye OEChem is not installed! See "
            "https://docs.eyesopen.com/toolkits/python/index.html for detail"
        )
    if not oechem_licensed:
        raise RuntimeError(
            "Cannot find a license for OpenEye OEChem! See "
            "https://docs.eyesopen.com/toolkits/python/index.html for detail"
        )
    if not oeomega_available:
        raise RuntimeError(
            "OpenEye OEOmega is not installed! See "
            "https://docs.eyesopen.com/toolkits/python/index.html for detail"
        )
    if not oeomega_licensed:
        raise RuntimeError(
            "Cannot find a license for OpenEye OEOmega! See "
            "https://docs.eyesopen.com/toolkits/python/index.html for detail"
        )


class SMILESMixin:
    """A mixin for handling SMILES."""

    @property
    def canonical_smiles(self):
        """Return the canonical SMILES string for this object."""
        return self.to_smiles(canonical=True)

    @property
    def smarts(self):
        """Return the SMARTS string for this object."""
        return self.to_smarts()

    @property
    def smiles(self):
        """Return the SMILES string for this object."""
        return self.to_smiles()

    def to_smiles(
        self,
        canonical=False,
        hydrogens=False,
        isomeric=True,
        flavor="rdkit",
    ):
        """Create the SMILES string from the system.

        Parameters
        ----------
        canonical : bool = False
            Whether to create canonical SMILES
        hydrogens : bool = False
            Whether to keep H's in the SMILES string.
        isomeric : bool = True
            Whether to use isomeric SMILES
        rdkit : bool = False
            Whether to use RDKit rather than default of OpenBabel

        Returns
        -------
        str
            The SMILES string, or (SMILES, name) if the rname is requested
        """
        logger.info("to_smiles")

        if flavor == "rdkit":
            mol = self.to_RDKMol()
            if isomeric:
                Chem.FindPotentialStereo(mol)
            if hydrogens:
                mol2 = mol
            else:
                mol2 = AllChem.RemoveHs(mol)
            if canonical:
                smiles = Chem.MolToSmiles(mol2, isomericSmiles=isomeric)
            else:
                smiles = Chem.MolToSmiles(
                    mol2, isomericSmiles=isomeric, canonical=False
                )
        elif flavor == "openbabel":
            obConversion = openbabel.OBConversion()
            if canonical:
                obConversion.SetOutFormat("can")
            else:
                obConversion.SetOutFormat("smi")

            mol = self.to_OBMol()

            if hydrogens:
                obConversion.AddOption("h")
            if not isomeric:
                obConversion.AddOption("i")
            smiles = obConversion.WriteString(mol)
        elif flavor == "openeye":
            check_openeye_license()

            mol = self.to_OEGraphMol()
            smiles = oechem.OECreateSmiString(mol)
        else:
            raise ValueError(f"flavor of SMILES '{flavor}' not supported")

        logger.info(f"smiles = '{smiles}'")

        return smiles.strip()

    def from_smiles(self, smiles, name=None, reorient=True, flavor="rdkit"):
        """Create the system from a SMILES string.

        Parameters
        ----------
        smiles : str
            The SMILES string
        name : str = None
            The name of the molecule
        reorient : bool = True
            Whether to reorient to the standard orientation
        rdkit : bool = False
            Whether to use RDKit rather than default of OpenBabel

        Returns
        -------
        None
        """

        save = self.name

        if flavor == "rdkit":
            mol = Chem.rdmolfiles.MolFromSmiles(smiles)
            if mol is None:
                raise ValueError(f"SMILES '{smiles}' is not valid.")
            mol = Chem.AddHs(mol)
            AllChem.EmbedMolecule(mol)
            self.from_RDKMol(mol)
        elif flavor == "openbabel":
            obConversion = openbabel.OBConversion()
            obConversion.SetInAndOutFormats("smi", "mdl")
            obConversion.AddOption("3")
            mol = openbabel.OBMol()
            obConversion.ReadString(mol, smiles)

            # Add hydrogens
            mol.AddHydrogens()

            # Get coordinates for a 3-D structure
            builder = openbabel.OBBuilder()
            builder.Build(mol)

            self.from_OBMol(mol)
        elif flavor == "openeye":
            check_openeye_license()
            mol = oechem.OEGraphMol()
            if not oechem.OESmilesToMol(mol, smiles):
                raise ValueError(f"SMILES '{smiles}' is not valid.")
            omega = oeomega.OEOmega()
            omega.SetMaxConfs(1)
            omega.SetStrictStereo(False)
            omega.SetStrictAtomTypes(False)

            # Add hydrogens
            oechem.OEAddExplicitHydrogens(mol)

            self.from_OEMol(mol)

        # Rotate to standard orientation
        if reorient:
            rdkMol = self.to_RDKMol()
            rdkConf = rdkMol.GetConformers()[0]
            Chem.rdMolTransforms.CanonicalizeConformer(rdkConf, ignoreHs=False)
            self.from_RDKMol(rdkMol)

        if name is not None:
            self.name = name
        else:
            self.name = save

    def to_smarts(self):
        """Generate a SMARTS string for this object.

        Returns
        -------
        str
            The SMARTS string.
        """

        generator = GenSMARTS(self)
        return generator.smarts


class GenSMARTS(object):
    """A class to generate SMARTS strings for an object.

    Parameters
    ----------
    mol_object : _Configuration, _Template, _Subset
    """

    def __init__(self, mol_object=None):
        self.smarts = ""
        self._mol_object = None
        self.symbols = []
        self.visited = []
        self.neighbors = []

        self.mol_object = mol_object

    @property
    def mol_object(self):
        """The molecular object to work with."""
        return self._mol_object

    @mol_object.setter
    def mol_object(self, value):
        self._mol_object = value

        if self._mol_object is not None:
            atoms = self.mol_object.atoms
            bonds = self.mol_object.bonds

            self.symbols = atoms.symbols

            # Find the neighbors of each atom for walking the structure
            tmp_neighbors = {i: [] for i in atoms.ids}
            for bond in bonds.bonds():
                i = bond["i"]
                j = bond["j"]
                bo = bond["bondorder"]
                tmp_neighbors[i].append((j, bo))
                tmp_neighbors[j].append((i, bo))

            # The index of each atom id in the list, 0..natoms-1
            index = {j: i for i, j in enumerate(atoms.ids)}

            # Convert to indices
            self.neighbors = {}
            for i, js in tmp_neighbors.items():
                self.neighbors[index[i]] = [(index[j], bo) for j, bo in js]

            # An array to keep track of the atoms visited
            self.visited = [False] * atoms.n_atoms

            # And generate the SMARTS string
            self.smarts = self._recurse(atom=0)

    def _recurse(self, atom):
        """Helper method for recursion to generate SMARTS.

        Parameters
        ----------
        atom : int
            Index of the atom to work with.
        """
        symbol = self.symbols[atom]
        self.visited[atom] = True

        # How many neighbors, and how many are hydrogens
        nX = len(self.neighbors[atom])
        nH = 0
        for neighbor, bondorder in self.neighbors[atom]:
            if self.symbols[neighbor] == "H":
                nH += 1

        # Find any non-hydrogen neighbors that have not been visited,
        # and recurse.
        sub_smarts = []
        for neighbor, bondorder in self.neighbors[atom]:
            if self.symbols[neighbor] != "H" and not self.visited[neighbor]:
                if bondorder == 1:
                    bond = ""
                elif bondorder == 2:
                    bond = "="
                elif bondorder == 3:
                    bond = "#"
                elif bondorder == 5:
                    # Aromatic
                    symbol = symbol.lower()
                    bond = ":"
                else:
                    bond = "~"
                sub_smarts.append(f"{bond}{self._recurse(neighbor)}")

        smarts = f"[{symbol}H{nH}X{nX}]"
        if len(sub_smarts) > 0:
            for sub in sub_smarts[0:-1]:
                smarts += f"({sub})"
            smarts += sub_smarts[-1]

        return smarts



# -*- coding: utf-8 -*-

import csv
import json
import logging
from pathlib import Path
import pkg_resources

logger = logging.getLogger(__name__)
standard_properties = {}


def add_properties_from_file(path):
    """The standard properties recognized by SEAMM.

    These are officially defined properties that can be used anywhere in SEAMM, as
    long as the type and definition correspond to the standard.

    Each property is defined by a string with up to three parts:

        <property name>#<code or 'experiment'>#<technique or model chemistry>

    The property name is required. In most cases this is followed by either 'experiment'
    or the name of the code, e.g. 'MOPAC', 'Gaussian', or 'VASP'. The final part,
    if present, is either the experimental technique used to measure the property, or
    the model chemistry, such as 'MP2/6-31G**', 'PM7', or a forcefield name such as
    'AMBER/ff19SB'.

    You can create other properties on the fly, but they follow the above convention
    and should have an appropriate code and, if necessary, model chemistry, so that
    they full name is unique and does not conflict with any other defined name.

    For example, the standard property "enthalpy of formation" refers to the
    experimental heat of formation, or a calculated value comparable to experimental
    values. If you are not sure what the heat of formation in e.g. MOPAC is, you could
    create a new property "enthalpy of formation#MOPAC#<parameterization>", which is
    clearly similar to the standard "enthalpy of formation". If the community decides
    that it is indeed the same, it can be replaced by the standard form, and also
    aliased to it for backwards compatibility.
    """
    global standard_properties
    with open(path, newline="", encoding="utf-8-sig") as fd:
        data = csv.reader(fd)
        line = 0
        for row in data:
            line += 1
            if line == 1:
                # Headers
                headers = [*row]
                if headers != [
                    "Property",
                    "Type",
                    "Units",
                    "Description",
                    "URL",
                ]:
                    raise ValueError(
                        "Header of standard properties file not valid: "
                        + ", ".join(headers)
                    )
            else:
                property = row[0]
                data = standard_properties[property] = {}
                for key, value in zip(headers[1:], row[1:]):
                    data[key] = value


path = Path(pkg_resources.resource_filename(__name__, "data/"))
csv_file = path / "standard_properties.csv"
add_properties_from_file(csv_file)


class _Properties(object):
    """A class for handling the properties in the star schema."""

    def __init__(self, system_db):
        self._system_db = system_db
        self._standard_properties = None

    @property
    def cursor(self):
        return self.system_db.cursor

    @property
    def db(self):
        return self.system_db.db

    @property
    def standard_properties(self):
        global standard_properties
        return standard_properties

    @property
    def system_db(self):
        """Return the SystemDB object."""
        return self._system_db

    def add(self, name, _type="float", units=None, description="", noerror=False):
        """Add a property to the database. By default, it is an error if it already
        exists!

        Parameters
        ----------
        name : str
            The name of the property.
        _type : str = "float"
            The type of the property -- float, int, str. Default is "float".
        units : str = None
            The units of the property value, as a string that Pint can interpret.
        description : str
            A longer, text description of the property.
        noerror: bool = False
            Whether to quietly ignore an existing entry

        Returns
        -------
        int
            The database id of the property.

        Note
        ----
        If the property is registered in the list of standard properties, just
        give the name and the rest will be filled in from the standard property
        metadata.
        """
        if self.exists(name):
            if noerror:
                return self.id(name)
            else:
                raise ValueError(f"Property '{name}' already exists.")

        table = self.system_db["property"]
        if name in self.standard_properties:
            data = self.standard_properties[name]
            result = table.append(
                name=name,
                type=data["Type"],
                units=data["Units"],
                description=data["Description"],
            )
        else:
            result = table.append(
                name=name,
                type=_type,
                units=units,
                description=description,
            )
        return result[0]

    def create_schema(self):
        """Add the needed tables to the database."""
        # The property dimension table
        table = self.system_db["property"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("name", coltype="str", index="unique")
        table.add_attribute("type", coltype="str")
        table.add_attribute("units", coltype="str")
        table.add_attribute("description", coltype="str")

        # Floating point facts
        table = self.system_db["float_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("system", coltype="int", references="system")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="float")

        self.db.execute(
            "CREATE INDEX float_data_idx_configuration_property_value"
            "    ON float_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_system_property_value"
            "    ON float_data(system, property, value)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_property_value"
            "    ON float_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_configuration_property"
            "    ON float_data(configuration, property)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_system_property"
            "    ON float_data(system, property)"
        )
        self.db.execute(
            "CREATE INDEX float_data_idx_configuration ON float_data(configuration)"
        )
        self.db.execute("CREATE INDEX float_data_idx_system ON float_data(system)")

        # Integer facts
        table = self.system_db["int_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("system", coltype="int", references="system")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="int")

        self.db.execute(
            "CREATE INDEX int_data_idx_configuration_property_value"
            "    ON int_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_system_property_value"
            "    ON int_data(system, property, value)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_property_value"
            "    ON int_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_configuration_property"
            "    ON int_data(configuration, property)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_system_property ON int_data(system, property)"
        )
        self.db.execute(
            "CREATE INDEX int_data_idx_configuration ON int_data(configuration)"
        )
        self.db.execute("CREATE INDEX int_data_idx_system ON int_data(system)")

        # String facts
        table = self.system_db["str_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("system", coltype="int", references="system")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="str")

        self.db.execute(
            "CREATE INDEX str_data_idx_configuration_property_value"
            "    ON str_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_system_property_value"
            "    ON str_data(system, property, value)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_property_value"
            "    ON str_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_configuration_property"
            "    ON str_data(configuration, property)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_system_property  ON str_data(system, property)"
        )
        self.db.execute(
            "CREATE INDEX str_data_idx_configuration ON str_data(configuration)"
        )
        self.db.execute("CREATE INDEX str_data_idx_system ON str_data(system)")

        # Array facts
        table = self.system_db["json_data"]
        table.add_attribute("id", coltype="int", pk=True)
        table.add_attribute("configuration", coltype="int", references="configuration")
        table.add_attribute("system", coltype="int", references="system")
        table.add_attribute("property", coltype="int", references="property")
        table.add_attribute("value", coltype="str")

        self.db.execute(
            "CREATE INDEX json_data_idx_configuration_property_value"
            "    ON json_data(configuration, property, value)"
        )
        self.db.execute(
            "CREATE INDEX json_data_idx_system_property_value"
            "    ON json_data(system, property, value)"
        )
        self.db.execute(
            "CREATE INDEX json_data_idx_property_value"
            "    ON json_data(property, value)"
        )
        self.db.execute(
            "CREATE INDEX json_data_idx_configuration_property"
            "    ON json_data(configuration, property)"
        )
        self.db.execute(
            "CREATE INDEX json_data_idx_system_property ON json_data(system, property)"
        )
        self.db.execute(
            "CREATE INDEX json_data_idx_configuration ON json_data(configuration)"
        )
        self.db.execute("CREATE INDEX json_data_idx_system ON json_data(system)")

    def description(self, _property):
        """The description of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The description of the property.
        """
        return self.metadata(_property)[2]

    def exists(self, name):
        """Whether the named property exists.

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        bool
            True if the property is registered, False otherwise.
        """
        self.cursor.execute("SELECT COUNT(*) FROM property WHERE name = ?", (name,))
        return self.cursor.fetchone()[0] != 0

    def get(self, configuration_id, _property="all", include_system_properties=False):
        """Get the given property value for the configuration.

        Parameters
        ----------
        configuration_id : int
            The id of the configuration.
        _property : int or str, or "all"
            The id or name of the property, or all properties if "all".
        include_system_properties : bool=False
            Whether to include properties that are on the system, not any configuration

        Returns
        -------
        int, float, or str
            The value of the property.
        """
        if include_system_properties:
            self.cursor.execute(
                "SELECT system FROM configuration WHERE id = ?", (configuration_id,)
            )
            system_id = self.cursor.fetchone()[0]

        if _property == "all":
            sql = (
                "SELECT name, type, value\n"
                "  FROM property, float_data\n"
                " WHERE float_data.property = property.id AND configuration = ?\n"
                " UNION \n"
                "SELECT name, type, value\n"
                "  FROM property, int_data\n"
                " WHERE int_data.property = property.id AND configuration = ?\n"
                " UNION \n"
                "SELECT name, type, value\n"
                "  FROM property, str_data\n"
                " WHERE str_data.property = property.id AND configuration = ?\n"
                " UNION \n"
                "SELECT name, type, value\n"
                "  FROM property, json_data\n"
                " WHERE json_data.property = property.id AND configuration = ?"
            )
            if include_system_properties:
                sql += (
                    "\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, float_data\n"
                    " WHERE float_data.property = property.id"
                    "   AND configuration IS NULL AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, int_data\n"
                    " WHERE int_data.property = property.id"
                    "   AND configuration IS NULL AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, str_data\n"
                    " WHERE str_data.property = property.id"
                    "   AND configuration IS NULL AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, json_data\n"
                    " WHERE json_data.property = property.id"
                    "   AND configuration IS NULL AND system = ? \n"
                )
                self.cursor.execute(
                    sql,
                    (
                        configuration_id,
                        configuration_id,
                        configuration_id,
                        configuration_id,
                        system_id,
                        system_id,
                        system_id,
                        system_id,
                    ),
                )
            else:
                self.cursor.execute(
                    sql,
                    (
                        configuration_id,
                        configuration_id,
                        configuration_id,
                        configuration_id,
                    ),
                )

            result = {}
            for row in self.cursor:
                name, _type, value = row
                if _type == "float":
                    result[name] = float(value)
                elif _type == "int":
                    result[name] = int(value)
                elif _type == "json":
                    result[name] = json.loads(value)
                else:
                    result[name] = value
            return result
        else:
            if isinstance(_property, str):
                if self.exists(_property):
                    pid = self.id(_property)
                else:
                    raise ValueError(f"Property '{_property}' does not exist.")
            else:
                pid = _property

            ptype = self.type(pid)
            sql = (
                f"SELECT value FROM {ptype}_data\n"
                "  WHERE configuration = ? AND property = ?"
            )
            if include_system_properties:
                sql += (
                    "\n"
                    "  UNION \n"
                    f"SELECT value FROM {ptype}_data\n"
                    "  WHERE configuration IS NULL and system = ? AND property = ?"
                )
                self.cursor.execute(sql, (configuration_id, pid, system_id, pid))
            else:
                self.cursor.execute(sql, (configuration_id, pid))
            result = self.cursor.fetchone()
            if result is None:
                raise ValueError(
                    f"Property {_property} does not exist for configuration "
                    f"{configuration_id}"
                )
            if ptype == "json":
                return json.loads(result[0])
            else:
                return result[0]

    def get_for_system(
        self, system_id, _property="all", include_configuration_properties=False
    ):
        """Get the given property value(s) for the system.

        Parameters
        ----------
        system_id : int
            The id of the system.
        _property : int or str of "all"
            The id or name of the property, or "all" for all properties.
        include_configuration_properties : bool=False
            Whether to include properties from any configuration in the system.

        Returns
        -------
        [int, float, or str]
            The value(s) of the property.
        """
        if _property == "all":
            if include_configuration_properties:
                sql = (
                    "SELECT name, type, value\n"
                    "  FROM property, float_data\n"
                    " WHERE float_data.property = property.id AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, int_data\n"
                    " WHERE int_data.property = property.id AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, str_data\n"
                    " WHERE str_data.property = property.id AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, json_data\n"
                    " WHERE json_data.property = property.id AND system = ?\n"
                )
            else:
                sql = (
                    "SELECT name, type, value\n"
                    "  FROM property, float_data\n"
                    " WHERE float_data.property = property.id"
                    "   AND configuration is NULL AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, int_data\n"
                    " WHERE int_data.property = property.id"
                    "   AND configuration is NULL AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, str_data\n"
                    " WHERE str_data.property = property.id"
                    "   AND configuration is NULL AND system = ?\n"
                    " UNION \n"
                    "SELECT name, type, value\n"
                    "  FROM property, json_data\n"
                    " WHERE json_data.property = property.id"
                    "   AND configuration is NULL AND system = ?\n"
                )
            self.cursor.execute(sql, (system_id, system_id, system_id, system_id))

            result = {}
            for row in self.cursor:
                name, _type, value = row
                if _type == "float":
                    result[name] = float(value)
                elif _type == "int":
                    result[name] = int(value)
                elif _type == "json":
                    result[name] = json.loads(value)
                else:
                    result[name] = value
        else:
            if isinstance(_property, str):
                if self.exists(_property):
                    pid = self.id(_property)
                else:
                    raise ValueError(f"Property '{_property}' does not exist.")
            else:
                pid = _property
            ptype = self.type(pid)

            if include_configuration_properties:
                sql = (
                    f"SELECT value FROM {ptype}_data WHERE system = ? AND property = ?"
                )
            else:
                sql = (
                    f"SELECT value FROM {ptype}_data"
                    "  WHERE configuration ISNULL AND system = ? AND property = ?"
                )
            result = []
            for row in self.db.execute(sql, (system_id, pid)):
                if ptype == "json":
                    result.append(json.loads(row[0]))
                else:
                    result.append(row[0])

        return result

    def id(self, name):
        """The id for a property

        Parameters
        ----------
        name : str
            The name of the property.

        Returns
        -------
        int
            The database id for the property.
        """
        self.cursor.execute("SELECT id FROM property WHERE name = ?", (name,))
        tmp = self.cursor.fetchone()
        if tmp is None:
            raise KeyError(f"Property '{name}' is not known.")
        else:
            return tmp[0]

    def known_properties(self):
        """List the known properties."""
        self.cursor.execute("SELECT name FROM property")
        return [row[0] for row in self.cursor.fetchall()]

    def metadata(self, _property):
        """The metadata for a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str, str, str
            The type, units, and description of the property
        """
        if isinstance(_property, str):
            self.cursor.execute(
                "SELECT type, units, description FROM property WHERE name = ?",
                (_property,),
            )
        else:
            self.cursor.execute(
                "SELECT type, units, description FROM property WHERE id = ?",
                (_property,),
            )
        tmp = self.cursor.fetchone()
        if tmp is not None:
            return tmp

        if _property in self.standard_properties:
            data = self.standard_properties[_property]
            return data["Type"], data["Units"], data["Description"]

        raise KeyError(f"Property '{_property}' is not known.")

    def name(self, pid):
        """The name of a property

        Parameters
        ----------
        pid : int
            The id of the property.

        Returns
        -------
        str
            The name of the property.
        """
        self.cursor.execute("SELECT name FROM property WHERE id = ?", (pid,))
        tmp = self.cursor.fetchone()
        if tmp is None:
            raise KeyError(f"Property id = '{pid}' is not known.")
        else:
            return tmp[0]

    def property_id(self, name):
        "Obsolete routine kept for compatibility"
        return self.id(name)

    def property_name(self, pid):
        "Obsolete routine kept for compatibility"
        return self.name(pid)

    def property_type(self, _property):
        "Obsolete routine kept for compatibility"
        return self.type(_property)

    def put(self, configuration_id, _property, value):
        """Store the given property value for the configuration.

        Parameters
        ----------
        configuration_id : int
            The id of the configuration.
        _property : int or str
            The id or name of the property.
        value : int, float, or str
            The value to store.
        """
        # Get the property id and type
        if isinstance(_property, str):
            if not self.exists(_property):
                if _property in self.standard_properties:
                    self.add(_property)
                else:
                    raise ValueError(f"Property '{_property}' does not exist.")
            pid = self.id(_property)
        else:
            pid = _property
        ptype = self.type(pid)

        if ptype == "json":
            value = json.dumps(value, separators=(",", ":"))

        # Get the system id
        self.cursor.execute(
            "SELECT system FROM configuration WHERE id = ?", (configuration_id,)
        )
        system_id = self.cursor.fetchone()[0]

        sql = (
            f"INSERT INTO {ptype}_data (configuration, system, property, value)"
            "      VALUES(?, ?, ?, ?)"
        )
        self.db.execute(sql, (configuration_id, system_id, pid, value))

    def put_for_system(self, system_id, _property, value):
        """Store the given property value for the system.

        Parameters
        ----------
        system_id : int
            The id of the system.
        _property : int or str
            The id or name of the property.
        value : int, float, or str
            The value to store.
        """
        # Get the property id and type
        if isinstance(_property, str):
            if not self.exists(_property):
                if _property in self.standard_properties:
                    self.add(_property)
                else:
                    raise ValueError(f"Property '{_property}' does not exist.")
            pid = self.id(_property)
        else:
            pid = _property
        ptype = self.type(pid)

        if ptype == "json":
            value = json.dumps(value, separators=(",", ":"))

        sql = (
            f"INSERT INTO {ptype}_data (system, property, value)"
            "      VALUES(?, ?, ?)"
        )
        self.db.execute(sql, (system_id, pid, value))

    def query(self, *args, what=["configuration"]):
        """Find configurations that match the query defined by the args.

        what : [str or int]]
            What to return ... "configuration", property names or ids.
        where : {}

        Note
        ----
        args should be a multiple of 3 in length, with each triplet being
        (property, operator, value). The property
        """
        sql = "SELECT"
        where = "WHERE"
        criteria = []
        tables = {}
        results = []
        types = []
        n_tables = 0
        for i, item in enumerate(what):
            if i > 0:
                sql += ","
            if isinstance(item, str):
                if item == "configuration":
                    sql += " t0.configuration"
                    results.append("configuration_id")
                    types.append["configuration"]
                else:
                    pid = self.id(item)
                    ptype = self.type(pid)
                    table = ptype + "_data"
                    if table not in tables:
                        tables[table] = {pid: f"t{n_tables}"}
                        n_tables += 1
                    elif pid not in tables[table]:
                        tables[table][pid] = f"t{n_tables}"
                        n_tables += 1
                    alias = tables[table][pid]
                    sql += f" {alias}.value"
                    criteria.append(f"{alias}.property == {pid}")
                    results.append(item)
                    types.append(ptype)
            elif isinstance(item, int):
                pid = item
                ptype = self.type(pid)
                table = ptype + "_data"
                if table not in tables:
                    tables[table] = {pid: f"t{n_tables}"}
                    n_tables += 1
                elif pid not in tables[table]:
                    tables[table][pid] = f"t{n_tables}"
                    n_tables += 1
                alias = tables[table][pid]
                sql += f" {alias}.value"
                criteria.append(f"{alias}.property == {pid}")
                results.append(self.name(pid))
                types.append(ptype)

        # And the WHERE clause ...
        where = {0: ""}
        items = iter(args)
        level = 0
        for item in items:
            if item == "(":
                level += 1
                where[level] = " ("
            elif item == ")":
                where[level] += ")"
                where[level - 1] += where[level]
                del where[level]
                level -= 1
            else:
                # Configuration or Property name or id
                if isinstance(item, str):
                    pid = self.id(item)
                else:
                    pid = item
                ptype = self.type(pid)
                table = ptype + "_data"
                if table not in tables:
                    tables[table] = {pid: f"t{n_tables}"}
                    n_tables += 1
                elif pid not in tables[table]:
                    tables[table][pid] = f"t{n_tables}"
                    n_tables += 1
                alias = tables[table][pid]
                if where[level] != "" and where[level] != "( ":
                    where[level] += " AND"
                where[level] += f" {alias}.property == {pid}"

                operator = next(items)
                if operator.lower() == "between":
                    if ptype == "str":
                        where[level] += (
                            f' AND {alias}.value BETWEEN "{next(items)}"'
                            f' AND "{next(items)}"'
                        )
                    if ptype == "json":
                        value1 = json.dumps(next(items), separator=(",", ":"))
                        value2 = json.dumps(next(items), separator=(",", ":"))
                        where[
                            level
                        ] += f' AND {alias}.value BETWEEN "{value1}" AND "{value2}"'
                    else:
                        where[level] += (
                            f" AND {alias}.value BETWEEN {next(items)}"
                            f" AND {next(items)}"
                        )
                else:
                    if ptype == "str":
                        where[level] += f' AND {alias}.value {operator} "{next(items)}"'
                    elif ptype == "json":
                        value = json.dumps(next(items), separator=(",", ":"))
                        where[level] += f' AND {alias}.value {operator} "{value}"'
                    else:
                        where[level] += f" AND {alias}.value {operator} {next(items)}"

        # Put it all together ... FROM clause plus needed criteria
        # if len(tables) == 0:
        #     # No tables ... so probably getting all configurations with properties
        #     tables = ["int_data", "float_data", "str_data"]
        sql += " FROM "

        tmp = []
        for table, data in tables.items():
            for pid, alias in data.items():
                tmp.append(f"{table} AS {alias}")
        sql += ", ".join(tmp)

        # Now for the WHERE clause
        sql += " WHERE "
        tmp = []
        for table, data in tables.items():
            for pid, alias in data.items():
                if alias != "t0":
                    tmp.append(f"{alias}.configuration = t0.configuration")
        sql += " AND ".join(tmp)

        if len(criteria) > 0:
            if len(tmp) > 0:
                sql += " AND "
            sql += " AND ".join(criteria)

        # The user criteria, if any
        if where[0] != "":
            if len(tmp) > 0:
                sql += " AND"
            sql += where[0]

        logger.info(sql)

        result = {item: [] for item in results}
        try:
            self.cursor.execute(sql)
        except Exception:
            print("")
            print(f"{tmp=}")
            print(f"{criteria=}")
            print(f"{where[0]=}")
            raise

        for row in self.cursor:
            for item, ptype, value in zip(results, types, row):
                if ptype == "json":
                    result[item].append(json.loads(value))
                else:
                    result[item].append(value)
        return result

    def type(self, _property):
        """The type of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The type of the property.
        """
        return self.metadata(_property)[0]

    def units(self, _property):
        """The unit string of a property

        Parameters
        ----------
        _property : int or str
            The id or name of the property.

        Returns
        -------
        str
            The units string for the property.
        """
        return self.metadata(_property)[1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the OpenEye mixin of the class."""

import pprint  # noqa: F401
import pytest  # noqa: F401

try:
    from openeye import oechem, oeomega
except ImportError:
    pass


@pytest.mark.openeye
def test_to_OEGraphMol(Acetate):
    """Test creating a OEGraphMol object from a structure."""
    correct = {
        "float property": 3.14,
        "float property,units": "kcal/mol",
        "int property": 2,
        "net charge": -1,
        "spin multiplicity": 1,
        "str property": "Hi!",
    }

    mol = Acetate.to_OEGraphMol(properties="all")

    assert Acetate.n_atoms == mol.NumAtoms()
    bond_list = [bond.GetOrder() for bond in mol.GetBonds()]
    assert Acetate.bonds.bondorders == bond_list

    data = {}
    for tmp in mol.GetDataIter():
        tag = tmp.GetTag()
        attribute = oechem.OEGetTag(tag)
        value = mol.GetData(tag)
        data[attribute] = value
    if data != correct:
        pprint.pprint(data)
    assert data == correct


@pytest.mark.openeye
def test_from_OEMol(configuration):
    """Test creating a structure from an RDKMol object."""
    correct = {"float property": 3.14, "integer property": 42, "string property": "foo"}

    mol = oechem.OEGraphMol()
    assert oechem.OESmilesToMol(mol, "CC=O")

    omega = oeomega.OEOmega()
    # Only generate one conformer for our molecule
    omega.SetMaxConfs(1)
    # Set to False to pick random stereoisomer if stereochemistry is not specified
    omega.SetStrictStereo(False)
    # Be a little loose about atom typing to ensure parameters are available to omega
    # for all molecules
    omega.SetStrictAtomTypes(False)

    # Add hydrogens
    oechem.OEAddExplicitHydrogens(mol, False, True)

    assert mol.NumAtoms() == 7

    # Add some properties
    mol.SetIntData("integer property", 42)
    mol.SetDoubleData("float property", 3.14)
    mol.SetStringData("string property", "foo")
    mol.SetIntData("net charge", -2)
    mol.SetIntData("spin multiplicity", 3)

    configuration.from_OEMol(mol)

    assert configuration.n_atoms == 7
    assert configuration.bonds.bondorders == [1, 2, 1, 1, 1, 1]

    result = configuration.properties.get()
    if result != correct:
        pprint.pprint(result)
    assert result == correct

    assert configuration.charge == -2
    assert configuration.spin_multiplicity == 3



# -*- coding: utf-8 -*-

from fractions import Fraction
import logging
import pprint  # noqa: F401
import re

import numpy as np
import spglib

logger = logging.getLogger(__name__)
# logger.setLevel("INFO")
logger.setLevel("WARNING")


class _Symmetry(object):
    """A class to handle point and space group symmetry."""

    spgno_to_hall = None
    spgname_to_hall = None
    spgname_to_system = None
    hall_to_spgname = None
    hall_to_hall_symbol = None
    hall_to_IT_symbol = None

    def __init__(self, configuration):
        """Initialize from the database.

        Parameters
        ----------
        system_db : SystemDB
            The SystemDB instance that we are working with.
        _id : int
            The id of this particular symmetry.
        """
        self._configuration = configuration
        self._system = self._configuration.system
        self._system_db = self._system.system_db
        self._id = configuration.symmetry_id
        self._operators = None  # 4x4 symmetry operation matrices
        self._symop_products = None  # Square matrix of symops that are products of 2
        self._atom_generators = None  # Sym operations that create the symmetric atoms
        self._symop_to_atom = None  # The symmetry atom resulting from symmetry ops
        self._atom_to_asymmetric_atom = None  # The asymmetric atom for each atom
        self._bond_to_asymmetric_bond = None  # The asymmetric bond for each bond
        self._bond_atoms = None  # The pair of symmetric atoms in each bond
        self._bond_offsets = None  # The triplet of cell offsets for each bond

        super().__init__()

    @staticmethod
    def filter_atoms(atoms, coordinates, group=None, sym_ops=None, operators=None):
        """Check and filter the atoms in a periodic system.

        There is an issue with some CIF files, notably those from the Cambridge
        structural database, where there are atoms in the asymmetric unit that
        are duplicates. It appears that the CIF file is written with entire molecules
        in some cases. So if the molecule is on a symmetry element, the atoms that are
        related by symmetry are explicitly in the file.

        To handle this we need to check for duplicates and remove them. Not a bad check
        to have, anyway.

        Parameters
        ----------
        atoms : str or int
            The atomic symbols or numbers of the atoms.
        coordinates : array_like
            The coordinates of the atoms.
        group : str
            The space group symbol.
        sym_ops : array_like
            The symmetry operations as strings.
        operators : array_like
            The symmetry operators.
        """
        if group is not None:
            # Get the symmetry operators
            if sym_ops is not None or operators is not None:
                raise RuntimeError(
                    "Cannot specify both group and sym_ops or operators."
                )
            operators = _Symmetry.get_operators(group)
        elif sym_ops is not None:
            if operators is not None or group is not None:
                raise RuntimeError(
                    "Cannot specify both sym_ops and group or operators."
                )
            operators = _Symmetry.get_operators(sym_ops)
        elif operators is not None:
            if sym_ops is not None or group is not None:
                raise RuntimeError(
                    "Cannot specify both operators and group or sym_ops."
                )
        else:
            raise RuntimeError("Must specify either group or sym_ops or operators.")

        self = _Symmetry()

        symbols = self.configuration.atoms.asymmetric_symbols
        n_atoms = self.configuration.n_asymmetric_atoms
        logger.info(f"Expanding {n_atoms} asymmetric atoms to full cell.")
        if n_atoms == 0:
            self._atom_generators = None
            self._symop_to_atom = None
        self._atom_generators = []  # Symmetry ops that create the symmetric atoms
        self._symop_to_atom = []  # The symmetry atom resulting from symmetry ops
        if self.configuration.periodicity == 3:
            if len(operators) == 1:
                # P1 is a special case. Nothing to do.
                self._atom_generators = [[0] for i in range(n_atoms)]
                self._symop_to_atom = [[i] for i in range(n_atoms)]
                self._atom_to_asymmetric_atom = [*range(n_atoms)]
            else:
                uvw0 = self.configuration.atoms.get_coordinates(
                    as_array=True, asymmetric=True
                )
                if uvw0.shape[0] != n_atoms:
                    raise RuntimeError(
                        f"Mismatch of number of atoms in symmetry: {uvw0.shape[0]} != "
                        f"{n_atoms}"
                    )
                logger.debug(
                    f"""
{n_atoms=}
{uvw0.shape=}"
Original coordinates
{uvw0}
                    """
                )

                uvw = np.ndarray((n_atoms, 4))
                uvw[:, 0:3] = uvw0[:, :]
                logger.debug(f"\nExpanded coordinates\n{uvw}")
                uvw[:, 3] = 1
                # logger.debug(self.symops)
                logger.debug(f"\n{operators.shape=}")
                # logger.debug(operators)
                # logger.debug(f"{uvw.shape=}")
                # logger.debug("Expanded coordinates")
                # logger.debug(uvw)
                xformed = np.einsum("ijk,lk", operators, uvw)
                logger.debug(f"\n{xformed.shape=}\n{xformed}")

                # For comparison, bring all atoms into the cell [0..1)
                tmp = xformed - np.floor(xformed)

                logger.debug(f"\ntmp = Transformed and floored coordinates\n{tmp}")

                # Keep track of atoms and coordinates found so can check for duplicates
                found = {}

                n_sym_atoms = 0
                n_asymmetric_atoms = 0
                for i in range(n_atoms):
                    values, I1, I2 = np.unique(
                        np.round(tmp[:, :3, i], 4),
                        axis=0,
                        return_index=True,
                        return_inverse=True,
                    )
                    # pprint.pprint(np.round(tmp[:, :3, i], 4).tolist())
                    logger.debug(
                        f"""
{i=}: {symbols[i]}
{np.round(tmp[:, :3, i], 4)}
nvalues
{values}
I1
{I1}
I2
{I2}
                        """
                    )
                    # Check for duplicates
                    if tuple(values[0]) in found:
                        if found[tuple(values[0])] != symbols[i]:
                            raise RuntimeError(
                                "Duplicate atoms with different symbols: "
                                f"{symbols[i]} != {found[tuple(values[0])]} at "
                                f"{values[0]}"
                            )
                        else:
                            # Same element, so ignore
                            logger.debug("\nDuplicate atom, ignoring")
                            continue

                    for value in values:
                        found[tuple(value)] = symbols[i]

                    n_asymmetric_atoms += 1
                    I2 += n_sym_atoms
                    self._atom_generators.append(I1.tolist())
                    self._symop_to_atom.append(I2.tolist())
                    n_sym_atoms += I1.shape[0]
                tmp = []
                for i, generators in enumerate(self._atom_generators):
                    tmp.extend([i] * len(generators))
                self._atom_to_asymmetric_atom = tmp
                logger.info(f"{n_asymmetric_atoms=} {n_sym_atoms=}")

    @staticmethod
    def get_operators(self, value, periodicity=3):
        """Get the symmetry operators as 4x4 matrices.

        Parameters
        ----------
        value : str or [str]
            The symmetry group or space group, or symmetry operators as strings.
        periodicity : int
            The periodicity of the structure. 3 for 3D, 2 for 2D, 1 for 1D, or 0.

        Returns
        -------
        operators : np.ndarray
            The symmetry operators as 4x4 matrices.
        """
        if isinstance(value, str):
            if value == "":
                raise RuntimeError("No group specified.")
            if self.periodicity == 0:
                if value != "C1":
                    raise NotImplementedError("Point groups not implemented yet!")
                W4 = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

                W4s = []
                W4s.append(W4)
                operators = np.array(W4s, dtype=float)
            else:
                # Get the 4x4 augmented matrices
                hall = _Symmetry.to_hall(value)
                value = _Symmetry.hall_to_spacegroup_name(hall)
                data = spglib.get_symmetry_from_database(hall)
                W4s = []
                for W, w in zip(
                    data["rotations"].tolist(), data["translations"].tolist()
                ):
                    W4 = []
                    for Wrow, w_i in zip(W, w):
                        W4.append([*Wrow, w_i])
                    W4.append([0, 0, 0, 1])
                    W4s.append(W4)
                operators = np.array(W4s, dtype=float)
        elif isinstance(value, list):
            W4s = []
            operators = np.array(W4s, dtype=float)
        return operators

    @staticmethod
    def hall_to_spacegroup_name(hall):
        """Return the International name including setting for a hall number."""
        if _Symmetry.hall_to_spgname is None:
            _Symmetry.spgname_to_hall = None
            _Symmetry.spacegroup_names_to_hall()
        return _Symmetry.hall_to_spgname[hall]

    @staticmethod
    def spacegroup_names_to_hall():
        """Dictionary of Hall number for spacegroup names."""
        if _Symmetry.spgname_to_hall is None:
            system_name = {
                "international_full": "name_H-M_full",
                "international": "name_H-M_alt",
                "international_short": "name_H-M_short",
                "hall_symbol": "name_Hall",
            }
            # Initialize the symmetry data
            _Symmetry.spgname_to_hall = {}
            _Symmetry.hall_to_spgname = {}
            _Symmetry.hall_to_hall_symbol = {}
            _Symmetry.hall_to_IT_number = {}
            if _Symmetry.spgname_to_system is None:
                _Symmetry.spgname_to_system = {}
            for hall in range(1, 530):
                data = spglib.get_spacegroup_type(hall)
                # pws pprint.pprint(data)
                choice = data["choice"]
                _Symmetry.hall_to_hall_symbol[hall] = data["hall_symbol"]
                _Symmetry.hall_to_IT_number[hall] = data["number"]

                # Handle Hall to spacegroup using the full H-M name
                key = "international_full"
                name = data[key]

                # Default setting if there are multiple, leave unadorned
                if choice in ("2", "b", "b1", "H"):
                    if (
                        hall in _Symmetry.hall_to_spgname
                        and name != _Symmetry.hall_to_spgname[hall]
                    ):
                        raise RuntimeError(
                            f"{hall=} {key} --> {name} exists: "
                            f"{_Symmetry.hall_to_spgname[hall]}"
                        )
                    else:
                        _Symmetry.hall_to_spgname[hall] = name
                else:
                    # Add other settings to the spacegroup name
                    if choice != "":
                        name += f" :{choice}"
                    if hall not in _Symmetry.hall_to_spgname:
                        # pws if choice != "":
                        # pws print(f"{hall} = {name}   QQQQQQQQQQQQQQQQQQQQQQ")
                        _Symmetry.hall_to_spgname[hall] = name

                # Now handle all the rest
                for key in (
                    "international_full",
                    "international",
                    "international_short",
                    "hall_symbol",
                ):
                    name = data[key]

                    # SPGLib encodes the international name: 'C 2/c = B 2/n 1 1',
                    if key == "international":
                        name = name.split("=")[0].strip()

                    # Default setting if there are multiple, leave unadorned
                    if choice in ("2", "b", "b1", "H"):
                        _Symmetry.spgname_to_hall[name] = hall
                        _Symmetry.spgname_to_system[name] = system_name[key]
                        for txt in ("_", " "):
                            tmp = name.replace(txt, "")
                            _Symmetry.spgname_to_hall[tmp] = hall
                            _Symmetry.spgname_to_system[tmp] = system_name[key]
                            if tmp[-2:] == ":H":
                                tmp = tmp[:-2].strip()
                                _Symmetry.spgname_to_hall[tmp] = hall
                                _Symmetry.spgname_to_system[tmp] = system_name[key]

                    if key in ("international", "international_short") and choice != "":
                        name += f" :{choice}"

                    _Symmetry.spgname_to_hall[name] = hall
                    _Symmetry.spgname_to_system[name] = system_name[key]
                    for txt in ("_", " "):
                        tmp = name.replace(txt, "")
                        _Symmetry.spgname_to_hall[tmp] = hall
                        _Symmetry.spgname_to_system[tmp] = system_name[key]
                        if tmp[-2:] == ":H":
                            tmp = tmp[:-2].strip()
                            _Symmetry.spgname_to_hall[tmp] = hall
                            _Symmetry.spgname_to_system[tmp] = system_name[key]
        return _Symmetry.spgname_to_hall

    @staticmethod
    def spacegroup_numbers_to_hall():
        """List of the Hall spacegroup names for the IT number."""
        if _Symmetry.spgno_to_hall is None:
            # Initialize the symmetry data
            _Symmetry.spgno_to_hall = {}
            if _Symmetry.spgname_to_system is None:
                _Symmetry.spgname_to_system = {}
            for hall in range(1, 530):
                data = spglib.get_spacegroup_type(hall)
                spgno = int(data["number"])
                if spgno not in _Symmetry.spgno_to_hall:
                    _Symmetry.spgno_to_hall[spgno] = hall
                    _Symmetry.spgname_to_system[spgno] = "Int_Tables_number"
                    _Symmetry.spgname_to_system[str(spgno)] = "Int_Tables_number"

        return _Symmetry.spgno_to_hall

    @staticmethod
    def to_hall(name):
        """Hall number given full spacegroup name or number."""
        if isinstance(name, int):
            return _Symmetry.spacegroup_numbers_to_hall()[name]
        return _Symmetry.spacegroup_names_to_hall()[name]

    @staticmethod
    def spacegroup_names_to_system():
        """Dictionary of system (name_Hall, name_H-M_full,...) for spacegroup names."""
        if _Symmetry.spgname_to_hall is None:
            _Symmetry.spacegroup_names_to_hall()
        if _Symmetry.spgno_to_hall is None:
            _Symmetry.spacegroup_numbers_to_hall()

        return _Symmetry.spgname_to_system

    @property
    def configuration(self):
        """Return the configuration."""
        return self._configuration

    @property
    def cursor(self):
        return self.system_db.cursor

    @property
    def db(self):
        return self.system_db.db

    @property
    def atom_generators(self):
        """The symmetry operations that create the symmetric atoms."""
        if self._atom_generators is None:
            self._expand()
        return self._atom_generators

    @property
    def atom_to_asymmetric_atom(self):
        """The asymmetric atom related to each atom."""
        if self._atom_to_asymmetric_atom is None:
            self._expand()
        return self._atom_to_asymmetric_atom

    @property
    def bonds_for_asymmetric_bonds(self):
        """List of bonds for each asymmetric bond."""
        if self.n_symops == 1:
            result = [[i] for i in range(self.configuration.bonds.n_asymmetric_bonds)]
        else:
            result = [[] for i in range(self.configuration.bonds.n_asymmetric_bonds)]
            to_asym = self.bond_to_asymmetric_bond
            for i, asym_bond in enumerate(to_asym):
                result[asym_bond].append(i)
        return result

    @property
    def bond_to_asymmetric_bond(self):
        """The asymmetric bond corresponding to each bond."""
        if self._bond_to_asymmetric_bond is None:
            self._expand_bonds()
        return self._bond_to_asymmetric_bond

    @property
    def bond_atoms(self):
        """The pair of symmetric atoms in each bond."""
        if self._bond_atoms is None:
            self._expand_bonds()
        return self._bond_atoms

    @property
    def bond_offsets(self):
        """The triplet of cell offsets of second atom in each bond."""
        if self._bond_offsets is None:
            self._expand_bonds()
        return self._bond_offsets

    @property
    def group(self):
        """The point or space group of the system"""
        self.cursor.execute('SELECT "group" FROM symmetry WHERE id = ?', (self.id,))
        return self.cursor.fetchone()[0]

    @group.setter
    def group(self, value):
        if value == "":
            self.Ws = None
            symops = None
        else:
            if self.configuration.periodicity == 0:
                if value != "C1":
                    raise NotImplementedError("Point groups not implemented yet!")
                symops = ["x,y,z"]
                W4 = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

                W4s = []
                W4s.append(W4)
                self._operators = np.array(W4s, dtype=float)
            else:
                # Get the 4x4 augmented matrices
                hall = self.to_hall(value)
                value = self.hall_to_spacegroup_name(hall)
                data = spglib.get_symmetry_from_database(hall)
                W4s = []
                for W, w in zip(
                    data["rotations"].tolist(), data["translations"].tolist()
                ):
                    W4 = []
                    for Wrow, w_i in zip(W, w):
                        W4.append([*Wrow, w_i])
                    W4.append([0, 0, 0, 1])
                    W4s.append(W4)
                self._operators = np.array(W4s, dtype=float)

                # The shorthand strings for the symmetry operations
                symops = []
                for W4 in W4s:
                    symop = []
                    for row in W4[0:3]:
                        line = ""
                        for r, xyz in zip(row[0:3], ("x", "y", "z")):
                            if r == 0:
                                pass
                            elif r == 1:
                                if line != "":
                                    line += "+"
                                line += xyz
                            elif r == -1:
                                line += "-" + xyz
                            else:
                                raise RuntimeError(f"bad rotation: '{row}'")
                        if row[3] == 0:
                            pass
                        else:
                            f = Fraction(row[3]).limit_denominator(10)
                            line += f"{f.numerator:+d}/{f.denominator}"
                        symop.append(line)
                    symops.append(",".join(symop))

        self.symops = symops
        self.db.execute(
            'UPDATE symmetry SET "group" = ? WHERE id = ?', (value, self.id)
        )
        self.db.commit()

    @property
    def hall_number(self):
        if self.configuration.periodicity == 0:
            return ""
        else:
            return self.to_hall(self.group)

    @property
    def hall_symbol(self):
        if self.configuration.periodicity == 0:
            return ""
        else:
            if _Symmetry.hall_to_hall_symbol is None:
                self.spacegroup_names_to_hall
            return _Symmetry.hall_to_hall_symbol[self.to_hall(self.group)]

    @property
    def id(self):
        """The id of this cell."""
        return self._id

    @property
    def inverse_operations(self):
        """The list of inverse symmetry operations for each symop."""
        return [x.index(0) for x in self.symop_products]

    @property
    def IT_number(self):
        if self.configuration.periodicity == 0:
            return ""
        else:
            if _Symmetry.hall_to_hall_symbol is None:
                self.spacegroup_names_to_hall
            return _Symmetry.hall_to_IT_number[self.to_hall(self.group)]

    @property
    def loglevel(self):
        """The logging level for this module."""
        result = logger.getEffectiveLevel()
        tmp = logging.getLevelName(result)
        if "Level" not in tmp:
            result = tmp
        return result

    @loglevel.setter
    def loglevel(self, value):
        logger.setLevel(value)

    @property
    def n_symops(self):
        """The number of symmetry operations."""
        return len(self.symops)

    @property
    def symmetry_matrices(self):
        """The symmetry operations as Numpy matrices (4 x 4 x n_ops)"""
        if self._operators is None:
            # Create the 4x4 matrices
            regexp = re.compile(
                r"([-+]?[0-9]/[0-9])?([-+]?[xyz][-+]?[xyz]?)([-+]?[0-9]/[0-9])?"
            )
            if len(self.symops) == 0:
                raise RuntimeError("The group is not defined!")
            W4s = []
            for line in self.symops:
                W4 = [
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0],
                ]
                for row, op in enumerate(line.split(",")):
                    op = op.strip()
                    match = regexp.fullmatch(op)
                    if match is None:
                        raise ValueError(
                            f"{op} is not a valid symmetry operation ({line})"
                        )
                    m1, m2, m3 = match.groups()
                    for column, xyz in enumerate(("x", "y", "z")):
                        if xyz in m2:
                            if "-" + xyz in m2:
                                W4[row][column] = -1
                            else:
                                W4[row][column] = 1
                    if m1 is None:
                        if m3 is None:
                            pass
                        else:
                            W4[row][3] = float(Fraction(m3))
                    else:
                        W4[row][3] = float(Fraction(m1))
                W4s.append(W4)
            self._operators = np.array(W4s, dtype=float)

        return self._operators

    @property
    def symop_products(self):
        """Matrix of symop ids that are the products of two other symops"""
        if self._symop_products is None:
            self._create_symop_products()
        return self._symop_products

    @property
    def symops(self):
        """The list of shorthand symmetry operations."""
        self.cursor.execute("SELECT symops FROM symmetry WHERE id = ?", (self.id,))
        text = self.cursor.fetchone()[0]

        result = []
        if text != "":
            result = text.split(" | ")
        return result

    @symops.setter
    def symops(self, ops):
        if ops is None:
            text = ""
        else:
            text = " | ".join(ops)
        # Ensure the lower case letters are used! x, y, z not X, Y, Z
        self.db.execute(
            "UPDATE symmetry SET symops = ? WHERE id = ?", (text.lower(), self.id)
        )
        # Unset the group
        self.db.execute('UPDATE symmetry SET "group" = "" WHERE id = ?', (self.id,))
        self.db.commit()

        self._operators = None
        self._symop_products = None
        self._atom_generators = None
        self._atom_to_asymmetric_atom = None
        self._symop_to_atom = None
        self._bond_to_asymmetric_bond = None
        self._bond_atoms = None
        self._bond_offsets = None

    @property
    def symop_to_atom(self):
        """The list of sym ops for each asymmetric atom to create symmetric atoms."""
        if self._symop_to_atom is None:
            self._expand()
        return self._symop_to_atom

    @property
    def system(self):
        """Return the System object that contains this cell."""
        return self._system

    @property
    def system_db(self):
        """Return the SystemDB object that contains this cell."""
        return self._system_db

    def find_spacegroup_from_operators(self):
        """Find the spacegroup from the symmetry operators."""
        if self.configuration.periodicity > 0:
            # Treat P1 as special to avoid bug in spglib...
            if self.n_symops == 1:
                hall_number = 1
                international_number = 1
            else:
                rotations = self.symmetry_matrices[:, 0:3, 0:3]
                translations = self.symmetry_matrices[:, 0:3, 3]
                data = spglib.get_spacegroup_type_from_symmetry(
                    rotations.tolist(),
                    translations.tolist(),
                    self.configuration.cell.vectors(),
                )
                hall_number = data["hall_number"]
                international_number = data["number"]
                while True:
                    data = self.configuration.get_symmetry_data(hall_number)
                    if data is None:
                        raise RuntimeError("Error finding spacegroup from operators.")
                    if data["number"] != international_number:
                        raise RuntimeError(
                            "Error finding setting for spacegroup number "
                            f"{international_number}."
                        )
                    if all(abs(data["origin_shift"]) < 0.001):
                        break
                    hall_number += 1
            return self.hall_to_spacegroup_name(hall_number)

    def reset_atoms(self):
        """The atoms have changed, so need to recalculate the symmetric atoms."""
        self._operators = None
        self._atom_generators = None
        self._atom_to_asymmetric_atom = None
        self._symop_to_atom = None

    def reset_bonds(self):
        """The bonds have changed, so need to recalculate the symmetric bonds."""
        self._bond_to_asymmetric_bond = None
        self._bond_atoms = None
        self._bond_offsets = None

    def symmetrize_atomic_scalar(self, v_sym):
        """Return the symmetrized scalars and deltas.

        Parameters
        ----------
        v_sym : [n_atoms * float] or numpy.ndarray
            The full set of scalar values for the symmetric atoms.

        Returns
        -------
        ([n_asymmetric_atoms * float] or numpy.ndarray, ditto)
            The asymmetric (symmetry unique) values in the same form as given,
            and the delta of the values from the average (or None for C1/P1)
        """
        if self.configuration.periodicity == 0:
            if self.n_symops == 1:
                return v_sym, None
            else:
                raise NotImplementedError("symmetrize_atomic_scalar for molecules!")

        if self.n_symops == 1:
            return v_sym, None

        v_in = np.array(v_sym)
        # print(f"{v_in=}")
        generators = self.atom_generators
        v = np.ndarray((len(generators),), dtype=float)
        delta = np.zeros_like(v_in)
        start = 0
        for asym_atom, ops in enumerate(generators):
            # print(f"{asym_atom=} {ops}")
            n = len(ops)
            # print(f"{n=}")
            v[asym_atom] = np.average(v_in[start : start + n], axis=0)
            delta[start : start + n] = v_in[start : start + n] - v[asym_atom]
            start += n

        if isinstance(v_sym, np.ndarray):
            return v, delta
        else:
            return v.tolist(), delta.tolist()

    def symmetrize_bond_scalar(self, v_sym):
        """Return the symmetrized scalar value and deltas for a bond property

        Parameters
        ----------
        v_sym : [n_bonds * float] or numpy.ndarray
            The full set of scalar values for the symmetric bonds.

        Returns
        -------
        ([n_asymmetric_bonds * float] or numpy.ndarray, ditto)
            The asymmetric (symmetry unique) values in the same form as given,
            and the delta of the values from the average (or None for C1/P1)
        """
        if self.configuration.periodicity == 0:
            if self.n_symops == 1:
                return v_sym, None
            else:
                raise NotImplementedError("symmetrize_bond_scalar for molecules!")

        if self.n_symops == 1:
            return v_sym, None

        v_in = np.array(v_sym)
        generators = self.bonds_for_asymmetric_bonds
        v = np.ndarray((len(generators),), dtype=float)
        delta = np.zeros_like(v_in)
        start = 0
        for asym_bond, ops in enumerate(generators):
            n = len(ops)
            v[asym_bond] = np.average(v_in[start : start + n], axis=0)
            delta[start : start + n] = v_in[start : start + n] - v[asym_bond]
            start += n

        if isinstance(v_sym, np.ndarray):
            return v, delta
        else:
            return v.tolist(), delta.tolist()

    def symmetrize_coordinates(self, xyz_sym, fractionals=True):
        """Return the symmetrized asymmetric coordinates and rms error.

        Parameters
        ----------
        xyz_sym : [n_atoms * [float]] or numpy.ndarray
            The full set of coordinates for the symmetric atoms.
        fractionals : bool = True
            Whether fractional (True) or Cartesian (False) coordinates are given.
            Only important for periodic systems, since molecules use Cartesians.

        Returns
        -------
        ([n_asymmetric_atoms * [float]] or numpy.ndarray, ditto)
            The asymmetric (symmetry unique) coordinates in the same form as given,
            and the delta of the atom positions from the average (or None for C1/P1)
        """
        if self.configuration.periodicity == 0:
            if self.n_symops == 1:
                return xyz_sym, None
            else:
                raise NotImplementedError("symmetrize_coordinates for molecules!")

        if self.n_symops == 1:
            return xyz_sym, None

        # We need fractional coordinates
        if fractionals:
            uvw_sym = np.array(xyz_sym)
        else:
            uvw_sym = self.configuration.cell.to_fractionals(xyz_sym, as_array=True)

        # Translate into cell
        uvw_sym -= np.floor(uvw_sym)

        generators = self.atom_generators
        inverse_ops = self.inverse_operations
        uvw = np.ndarray((len(generators), 3), dtype=float)
        delta = np.zeros_like(uvw_sym)
        atom = 0
        for asym_atom, ops in enumerate(generators):
            tmp = np.ndarray((len(ops), 3), dtype=float)
            start = atom
            for i, op in enumerate(ops):
                tmp[i] = self.vector_x_symop(uvw_sym[atom], inverse_ops[op])
                atom += 1
            tmp -= np.floor(tmp)
            uvw[asym_atom] = np.average(tmp, axis=0)
            delta[start:atom] = tmp - uvw[asym_atom]

        if isinstance(xyz_sym, np.ndarray):
            if fractionals:
                return uvw.round(8), delta.round(8)
            else:
                tmp = self.configuration.cell.to_cartesians(uvw, as_array=True)
                return tmp.round(8), delta.round(8)
        else:
            if fractionals:
                return uvw.round(8).tolist(), delta.round(8).tolist()
            else:
                tmp = self.configuration.cell.to_cartesians(uvw, as_array=True)
                return tmp.round(8).tolist(), delta.round(8).tolist()

    def update_group(self, value):
        if self.configuration.periodicity == 0:
            if value != "C1":
                raise NotImplementedError("Point groups not implemented yet!")
        self.db.execute(
            'UPDATE symmetry SET "group" = ? WHERE id = ?', (value, self.id)
        )
        self.db.commit()

    def vector_x_symop(self, vector, symop, translation=True):
        """Multiply a vector by a symmetry matrix."""
        sym_mat = self.symmetry_matrices[symop]
        if translation:
            v = np.array([0.0, 0.0, 0.0, 1.0])
        else:
            v = np.array([0.0, 0.0, 0.0, 0.0])
        v[0:3] = vector
        xformed = np.einsum("ij,j", sym_mat, v)
        return xformed[0:3]

    def _expand(self):
        """Setup the information for going from asymmetric cell to full cell.

        There is an issue with some CIF files, notably those from the Cambridge
        structural database, where there are atoms in the asymmetric unit that
        are duplicates. It appears that the CIF file is written with entire molecules
        in some cases. So if the molecule is on a symmetry element, the atoms that are
        related by symmetry are explicitly in the file.

        To handle this we need to check for duplicates and remove them. Not a bad check
        to have, anyway.
        """
        operators = self.symmetry_matrices

        symbols = self.configuration.atoms.asymmetric_symbols
        n_atoms = self.configuration.n_asymmetric_atoms
        logger.info(f"Expanding {n_atoms} asymmetric atoms to full cell.")
        if n_atoms == 0:
            self._atom_generators = None
            self._symop_to_atom = None
        self._atom_generators = []  # Symmetry ops that create the symmetric atoms
        self._symop_to_atom = []  # The symmetry atom resulting from symmetry ops
        if self.configuration.periodicity == 3:
            if len(operators) == 1:
                # P1 is a special case. Nothing to do.
                self._atom_generators = [[0] for i in range(n_atoms)]
                self._symop_to_atom = [[i] for i in range(n_atoms)]
                self._atom_to_asymmetric_atom = [*range(n_atoms)]
            else:
                uvw0 = self.configuration.atoms.get_coordinates(
                    as_array=True, asymmetric=True
                )
                if uvw0.shape[0] != n_atoms:
                    raise RuntimeError(
                        f"Mismatch of number of atoms in symmetry: {uvw0.shape[0]} != "
                        f"{n_atoms}"
                    )
                logger.debug(
                    f"""
{n_atoms=}
{uvw0.shape=}"
Original coordinates
{uvw0}
                    """
                )

                uvw = np.ndarray((n_atoms, 4))
                uvw[:, 0:3] = uvw0[:, :]
                logger.debug(f"\nExpanded coordinates\n{uvw}")
                uvw[:, 3] = 1
                # logger.debug(self.symops)
                logger.debug(f"\n{operators.shape=}")
                # logger.debug(operators)
                # logger.debug(f"{uvw.shape=}")
                # logger.debug("Expanded coordinates")
                # logger.debug(uvw)
                xformed = np.einsum("ijk,lk", operators, uvw)
                logger.debug(f"\n{xformed.shape=}\n{xformed}")

                # For comparison, bring all atoms into the cell [0..1)
                tmp = xformed - np.floor(xformed)

                logger.debug(f"\ntmp = Transformed and floored coordinates\n{tmp}")

                # Keep track of atoms and coordinates found so can check for duplicates
                found = {}

                n_sym_atoms = 0
                n_asymmetric_atoms = 0
                for i in range(n_atoms):
                    values, I1, I2 = np.unique(
                        np.round(tmp[:, :3, i], 4),
                        axis=0,
                        return_index=True,
                        return_inverse=True,
                    )
                    # pprint.pprint(np.round(tmp[:, :3, i], 4).tolist())
                    logger.debug(
                        f"""
{i=}: {symbols[i]}
{np.round(tmp[:, :3, i], 4)}
nvalues
{values}
I1
{I1}
I2
{I2}
                        """
                    )
                    # Check for duplicates
                    if tuple(values[0]) in found:
                        if found[tuple(values[0])] != symbols[i]:
                            raise RuntimeError(
                                "Duplicate atoms with different symbols: "
                                f"{symbols[i]} != {found[tuple(values[0])]} at "
                                f"{values[0]}"
                            )
                        else:
                            # Same element, so ignore
                            logger.debug("\nDuplicate atom, ignoring")
                            continue

                    for value in values:
                        found[tuple(value)] = symbols[i]

                    n_asymmetric_atoms += 1
                    I2 += n_sym_atoms
                    self._atom_generators.append(I1.tolist())
                    self._symop_to_atom.append(I2.tolist())
                    n_sym_atoms += I1.shape[0]
                tmp = []
                for i, generators in enumerate(self._atom_generators):
                    tmp.extend([i] * len(generators))
                self._atom_to_asymmetric_atom = tmp
                logger.info(f"{n_asymmetric_atoms=} {n_sym_atoms=}")

    def _expand_bonds(self):
        """Expand the list of asymmetric bonds to the full list.

        Keep track of the following:

            1. bond_to_asymmetric_bond: the asymmetric bond index for each bond
            2. bond_atoms: The pairs of symmetric atoms that form the bond
            3. bond_offsets: Triplets of offsets for bonds
        """
        logger.debug("In expand_bonds")

        if self._bond_to_asymmetric_bond is not None:
            return

        self._bond_to_asymmetric_bond = []
        self._bond_atoms = []
        self._bond_offsets = []

        bonds = self.configuration.bonds
        if bonds.n_asymmetric_bonds == 0:
            return

        indx = {j: i for i, j in enumerate(self.configuration.atoms.ids)}

        Is = [indx[i] for i in bonds.get_column_data("i")]
        Js = [indx[j] for j in bonds.get_column_data("j")]
        symop1s = bonds.get_column_data("symop1")
        symop2s = bonds.get_column_data("symop2")

        symop_to_atom = self.symop_to_atom
        product = self.symop_products
        n_symops = self.n_symops

        # Need to coordinates to sort out offsets introduced by symmetry
        uvw = self.configuration.atoms.get_coordinates(as_array=True)
        uvw_sym = self.configuration.atoms.get_coordinates(
            as_array=True, asymmetric=True
        )

        logger.debug(f"Asymmetric Coordinates:\n{str(uvw_sym)}\n\n")
        logger.debug(f"Coordinates:\n{str(uvw)}\n\n")

        asym_bonds = {}
        found = []
        # First work out offsets and look for bonds sharing atoms but different offsets
        for i, j, symop1, symop2 in zip(Is, Js, symop1s, symop2s):
            if symop1 == ".":
                op1 = 0
                offset1 = [0, 0, 0]
            else:
                if "_" in symop1:
                    op1, tmp = symop1.split("_")
                    op1 = int(op1) - 1
                    offset1 = [int(q) - 5 for q in tmp]
                else:
                    op1 = int(symop1) - 1
                    offset1 = [0, 0, 0]
            if symop2 == ".":
                op2 = 0
                offset2 = [0, 0, 0]
            else:
                if "_" in symop2:
                    op2, tmp = symop2.split("_")
                    op2 = int(op2) - 1
                    offset2 = [int(q) - 5 for q in tmp]
                else:
                    op2 = int(symop2) - 1
                    offset2 = [0, 0, 0]

            # Check if we've seen this pair, but with different offset.
            if i < j:
                key = (i, op1, j, op2)
                offkey = (offset1, offset2)
            else:
                key = (j, op2, i, op1)
                offkey = (offset2, offset1)
            logger.debug(f"{key=} {offkey}")
            if key in asym_bonds:
                if offkey in asym_bonds[key]:
                    raise RuntimeError(
                        f"Duplicate bonds specified for {i} ({symop1}) --  "
                        f"{j} ({symop2})"
                    )
                else:
                    asym_bonds[key].append(offkey)
            else:
                asym_bonds[key] = [offkey]

        bond_no = -1
        for ij, offsets in asym_bonds.items():
            i, op1, j, op2 = ij
            use_offsets = len(offsets) > 1 or (i == j and op1 == op2)
            for tmp in offsets:
                bond_no += 1
                offset1, offset2 = tmp
                if use_offsets:
                    logger.debug(
                        f"bond {bond_no}: {i}({op1}) - {j}({op2}) {offset1} {offset2}"
                    )
                else:
                    logger.debug(
                        f"bond {bond_no}: {i}({op1}) - {j}({op2}) ignore offsets"
                    )

                for op in range(n_symops):
                    prod_op1 = product[op1][op]
                    prod_op2 = product[op2][op]
                    iatom = symop_to_atom[i][prod_op1]
                    jatom = symop_to_atom[j][prod_op2]
                    ioff = self.vector_x_symop(offset1, prod_op1, translation=False)
                    joff = self.vector_x_symop(offset2, prod_op2, translation=False)

                    if iatom == jatom:
                        logger.debug(
                            f"iatom == jatom ({iatom}, {jatom}) {use_offsets=}"
                        )
                        if use_offsets:
                            if np.all(offset1 == offset2):
                                logger.debug(f"{offset1=} {offset2=}")
                                continue
                        else:
                            continue

                    if iatom > jatom:
                        ii, jj = j, i
                        iatom, jatom = jatom, iatom
                        offset1, offset2 = offset2, offset1
                        prod_op1, prod_op2 = prod_op2, prod_op1
                    else:
                        ii, jj = i, j

                    uvw1 = self.vector_x_symop(uvw_sym[ii], prod_op1)
                    uvw2 = self.vector_x_symop(uvw_sym[jj], prod_op2)

                    delta = uvw2 - uvw1
                    delta = delta.round(4)
                    # Not sure is <= or < :-)
                    tmpoff = np.select([delta < -0.5, delta > 0.5], [1, -1], 0)
                    if use_offsets:
                        ioff = np.array(offset1)
                        joff = np.array(tmpoff) + offset2
                    else:
                        ioff = np.array((0, 0, 0))
                        joff = np.array(tmpoff)

                    delta += joff - ioff

                    if np.any(abs(delta) > 1.0):
                        logger.debug(f"Throwing out delta = {str(delta)}")
                        continue

                    if (iatom, jatom, delta.tolist()) in found:
                        continue

                    found.append((iatom, jatom, delta.tolist()))

                    self._bond_to_asymmetric_bond.append(bond_no)
                    self._bond_atoms.append((iatom, jatom))
                    self._bond_offsets.append((ioff.tolist(), joff.tolist()))

                    if logger.isEnabledFor(logging.DEBUG):
                        uvw1 = self.vector_x_symop(uvw_sym[ii], prod_op1)
                        uvw2 = self.vector_x_symop(uvw_sym[jj], prod_op2)

                        delta = uvw2 - uvw1
                        tmpoff = np.select([delta <= -0.5, delta > 0.5], [1, -1], 0)
                        delta += joff - ioff
                        delta_xyz = self.configuration.cell.to_cartesians(
                            delta, as_array=True
                        )
                        r = np.round(np.linalg.norm(delta_xyz), 4)
                        xi, yi, zi = uvw1.tolist()
                        xj, yj, zj = uvw2.tolist()
                        xd, yd, zd = delta.tolist()
                        off1x, off1y, off1z = ioff.tolist()
                        off2x, off2y, off2z = joff.tolist()
                        logger.debug(
                            f"\t{iatom:3} {jatom:3} | {ii} {product[op1][op]:3} "
                            f"{jj} {product[op2][op]:3} | "
                            f"{xi:5.2f} {yi:5.2f} {zi:5.2f}"
                            f"({off1x:4} {off1y:4} {off1z:4}) "
                            f"{xj:5.2f} {yj:5.2f} {zj:5.2f}"
                            f"({off2x:4} {off2y:4} {off2z:4}) "
                            f"| {xd:5.2f} {yd:5.2f} {zd:5.2f}"
                            f" |  {r=:7.4f}"
                        )

    def _create_symop_products(self):
        """Create the matrix of the product of symmetry operators."""
        products = []
        ops = self.symmetry_matrices
        n = self.n_symops
        # Take the i'th symop and multiply by all the rest
        for i in range(n):
            prod = np.einsum("jk,ikl", ops[i, :, :], ops)
            row = []
            # For the i-j product, find it in the original list
            for j in range(n):
                # logger.debug(f"{j}:")

                # Subtract the product from the original symops
                tmp = ops - prod[j, :, :]

                # Shift any translation into the cell, i.e. [0,1)
                tmp[:, :, 3] = tmp[:, :, 3] - np.floor(tmp[:, :, 3])

                # We are looking for zeros, so sum the abs values
                tmp = tmp.reshape(n, 16)
                tmp2 = np.sum(np.abs(tmp), axis=1)

                # And find any matrices that are all zero
                hits = np.where(tmp2 == 0.0)[0]

                if len(hits) != 1:
                    symops = self.symops
                    logger.warning("Symmetry matrices")
                    for k in range(n):
                        logger.warning(f"{k}: {symops[k]}\n{str(ops[k])}")
                    logger.warning("")
                    logger.warning(f"Symop {i}:\n{str(ops[i])}")
                    logger.warning("")
                    logger.warning("Products")
                    for k in range(n):
                        logger.warning(f"\n{str(tmp[k])}")
                    raise RuntimeError(f"Problem with products of symops: {len(hits)=}")
                k = hits[0]
                row.append(k)
                # logger.debug(prod[j])
                # logger.debug(f"symop {k} is")
                # logger.debug(ops[k])
            if len(row) != n:
                raise RuntimeError(f"Not enough products of the symop {i}")
            # logger.debug(f"{i}: ({len(row)}) {row}")
            products.append(row)
        self._symop_products = products
